# -*- coding: utf-8 -*-
"""
High level functions to make complex tasks easier.
"""
import os as _os

###############################################################################
#                               Import Ourself                                #
###############################################################################

from . import run as _run
from . import conf as _conf
from . import logme as _logme
from . import options as _options
from .job import Job as _Job

###############################################################################
#                       Try Import Non-Required Modules                       #
###############################################################################

try:
    import numpy as _np
    import pandas as _pd
except ImportError:
    _logme.log('Could not import numpy and pandas for helpers', 'debug')

__all__ = ['parapply', 'parapply_summary', 'splitrun']

###############################################################################
#                                  parapply                                   #
###############################################################################


def parapply_summary(jobs, df, func, args=(), profile=None, applymap=False,
                     name='parapply', imports=None, direct=True, **kwds):
    """Run parapply for a function with summary stats.

    Instead of returning the concatenated result, merge the result using
    the same function as was used during apply.

    This works best for summary functions like `.mean()`, which do a linear
    operation on a whole dataframe or series.

    Args:
        jobs (int):         Number of pieces to split the dataframe into
        df (DataFrame):     Any pandas DataFrame
        args (tuple):       Positional arguments to pass to the function,
                            keyword arguments can just be passed directly.
        profile (str):      A fyrd cluster profile to use
        applymap (bool):    Run applymap() instead of apply()
        merge_axis (int):   Which axis to merge on, 0 or 1, default is 1 as
                            apply transposes columns
        merge_apply (bool): Apply the function on the merged dataframe also
        name (str):         A prefix name for all of the jobs
        imports (list):     A list of imports in any format, e.g.
                            ['import numpy', 'scipy', 'from numpy import mean']
        direct (bool):      Whether to run the function directly or to
                            return a Job. Default True.

        Any keyword arguments recognized by fyrd will be used for job
        submission.

        *Additional keyword arguments will be passed to DataFrame.apply()*

    Returns:
        DataFrame: A recombined DataFrame
    """
    if direct:
        return _parapply_summary(
            jobs, df, func, args=args, profile=profile, applymap=applymap,
            name=name, imports=imports, **kwds
        )
    else:
        kwargs = dict(
            args=args, profile=profile, applymap=applymap,
            name=name, imports=imports
            )
        kwds = _options.sanitize_arguments(kwds)
        kwargs.update(kwds)
        kwargs['imports']  = _run.get_all_imports(func, kwargs)
        kwargs['syspaths'] = _run.update_syspaths(func, kwargs)
        return _wrap_runner(
            _parapply_summary,
            *(jobs, df, func),
            **kwargs
        )


def _parapply_summary(jobs, df, func, args=(), profile=None, applymap=False,
                      name='parapply', imports=None, **kwds):
    """Direct running function for parapply_sumary, see that docstring."""
    imports = _run.export_imports(func, {'imports': imports})
    out = parapply(jobs, df, func, args, profile, applymap, name=name,
                   merge_axis=1, imports=imports, **kwds)

    # Pick function and get args
    sub_func = _run_applymap if applymap else _run_apply
    pandas_kwds = _options.split_keywords(kwds)[1]

    # Transpose
    out = out.T

    # Run the function again
    return sub_func(out, func, args, pandas_kwds)


def parapply(jobs, df, func, args=(), profile=None, applymap=False,
             merge_axis=0, merge_apply=False, name='parapply', imports=None,
             direct=True, **kwds):
    """Split a dataframe, run apply in parallel, return result.

    This function will split a dataframe into however many pieces are requested
    with the jobs argument, run apply in parallel by submitting the jobs to the
    cluster, and then recombine the outputs.

    If the 'clean_files' and 'clean_outputs' arguments are not passed, we
    delete all intermediate files and output files by default.

    This function will take any keyword arguments accepted by Job, which can
    be found by running fyrd.options.option_help(). It also accepts any of
    the keywords accepted by by pandas.DataFrame.apply(), found
    `here <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.apply.html>`_

    Args:
        jobs (int):         Number of pieces to split the dataframe into
        df (DataFrame):     Any pandas DataFrame
        args (tuple):       Positional arguments to pass to the function,
                            keyword arguments can just be passed directly.
        profile (str):      A fyrd cluster profile to use
        applymap (bool):    Run applymap() instead of apply()
        merge_axis (int):   Which axis to merge on, 0 or 1, default is 1 as
                            apply transposes columns
        merge_apply (bool): Apply the function on the merged dataframe also
        name (str):         A prefix name for all of the jobs
        imports (list):     A list of imports in any format, e.g.
                            ['import numpy', 'scipy', 'from numpy import mean']
        direct (bool):      Whether to run the function directly or to
                            return a Job. Default True.

        Any keyword arguments recognized by fyrd will be used for job
        submission.

        *Additional keyword arguments will be passed to DataFrame.apply()*

    Returns:
        DataFrame: A recombined DataFrame
    """
    if direct:
        return _parapply(
            jobs, df, func, args=args, profile=profile, applymap=applymap,
            merge_axis=merge_axis, merge_apply=merge_apply, name=name,
            imports=imports, **kwds
        )
    else:
        kwargs = dict(
            args=args, profile=profile, applymap=applymap,
            merge_axis=merge_axis, merge_apply=merge_apply, name=name,
            imports=imports
        )
        kwds = _options.sanitize_arguments(kwds)
        kwargs.update(kwds)
        kwargs['imports']  = _run.get_all_imports(func, kwargs)
        kwargs['syspaths'] = _run.update_syspaths(func, kwargs)
        return _wrap_runner(
            _parapply,
            *(jobs, df, func),
            **kwargs
        )


def _parapply(jobs, df, func, args=(), profile=None, applymap=False,
              merge_axis=0, merge_apply=False, name='parapply', imports=None,
              **kwds):
    """Direct running function for parapply, see parapply docstring."""
    # Handle arguments
    if not isinstance(jobs, int):
        raise ValueError('Jobs argument must be an integer.')
    if not isinstance(df, (_pd.core.frame.DataFrame, _np.ndarray)):
        raise ValueError('df must be a dataframe or numpy array, is {}'
                         .format(type(df)))
    if not callable(func):
        raise ValueError('function must be callable, current type is {}'
                         .format(type(func)))
    if profile is not None and not isinstance(profile, str):
        raise ValueError('Profile must be a string, is {}'
                         .format(type(profile)))
    fyrd_kwds, pandas_kwds = _options.split_keywords(kwds)

    # Get name
    name = name if name else 'split_file'

    # Set up auto-cleaning
    if 'clean_files' not in fyrd_kwds:
        fyrd_kwds['clean_files'] = True
    if 'clean_outputs' not in fyrd_kwds:
        fyrd_kwds['clean_outputs'] = True

    # Split dataframe
    _logme.log('Splitting dataframe', 'debug')
    dfs = _np.array_split(df, jobs)
    assert len(dfs) == jobs

    # Pick function
    sub_func = _run_applymap if applymap else _run_apply

    # Some sane imports
    imports  = _run.listify(imports) if imports else []
    imports += ['import numpy as np', 'import numpy', 'import scipy as sp',
                'import sp', 'import pandas as pd', 'import pandas',
                'from matplotlib import pyplot as plt',
                'from scipy import stats']
    imports = _run.export_imports(func, {'imports': imports})

    # Run the functions
    _logme.log('Submitting jobs', 'debug')
    outs = []
    count = 1
    for d in dfs:
        nm = '{}_{}_of_{}'.format(name, count, jobs)
        outs.append(
            _Job(sub_func, (d, func, args, pandas_kwds),
                 profile=profile, name=nm, imports=imports,
                 **fyrd_kwds).submit()
        )
        count += 1

    # Get the results
    _logme.log('Waiting for results', 'debug')
    results = []
    for out in outs:
        try:
            results.append(out.get())
        except IOError:
            _logme.log('Result getting failed, most likely one of the child ' +
                       'jobs crashed, check the error files', 'critical')
            raise

    # Return the recombined DataFrame
    _logme.log('Done, joinging', 'debug')
    try:
        out = _pd.concat(results, axis=merge_axis)
    except ValueError:
        _logme.log('DataFrame concatenation failed!', 'critical')
        raise

    if merge_apply:
        out = sub_func(out, func, args, pandas_kwds)

    return out


def _run_apply(df, func, args=None, pandas_kwds=None):
    """Run DataFrame.apply().

    Args:
        df (DataFrame):     Any pandas DataFrame
        args (tuple):       A tuple of arguments to submit to the function
        pandas_kwds (dict): A dictionary of keyword arguments to pass to
                            DataFrame.apply()

    Returns:
        DataFrame: The result of apply()
    """
    apply_kwds = {'args': args} if args else {}
    if pandas_kwds:
        apply_kwds.update(pandas_kwds)
    return df.apply(func, **apply_kwds)


def _run_applymap(df, func, args=None, pandas_kwds=None):
    """Run DataFrame.applymap().

    Args:
        df (DataFrame):     Any pandas DataFrame
        args (tuple):       A tuple of arguments to submit to the function
        pandas_kwds (dict): A dictionary of keyword arguments to pass to
                            DataFrame.apply()

    Returns:
        DataFrame: The result of apply()
    """
    apply_kwds = {'args': args} if args else {}
    if pandas_kwds:
        apply_kwds.update(pandas_kwds)
    return df.applymap(func, **apply_kwds)


###############################################################################
#                                  Split Run                                  #
###############################################################################


def splitrun(jobs, infile, inheader, command, args=None, kwargs=None,
             name=None, qtype=None, profile=None, outfile=None,
             outheader=False, merge_func=None, direct=True, **kwds):
    """Split a file, run command in parallel, return result.

    This function will split a file into however many pieces are requested
    with the jobs argument, and run command on each.

    Accepts exactly the same arguments as the Job class, with the exception of
    the first three and last four arguments, which are::

        - the number of jobs
        - the file to work on
        - whether the input file has a header
        - an optional output file
        - whether the output file has a header
        - an optional function to use to merge the resulting list, only used
          if there is no outfile.
        - whether to run directly or to return a Job. If direct is True, this
          function will just run and thus block until complete, if direct is
          False, the function will submit as a Job and return that Job.

    **Note**: If command is a string, `.format(file={file})` will be called on
    it, where file is each split file. If command is a function, the there must
    be an argument in either args or kwargs that contains `{file}`. It will be
    replaced with the *path to the file*, again by the format command.

    If outfile is specified, there must also be an '{outfile}' line in any
    script or an '{outfile}' argument in either args or kwargs. When this
    function completes, the file at outfile will contain the concatenated
    output files of all of the jobs.

    If the 'clean_files' and 'clean_outputs' arguments are not passed, we
    delete all intermediate files and output files by default.

    The intermediate files will be stored in the 'scriptpath' directory.

    Any header line is kept at the top of the file.

    Args:
        jobs (int):             Number of pieces to split the dataframe into
        infile (str):           The path to the file to be split.
        inheader (bool):        Does the input file have a header?
        command (function/str): The command or function to execute.
        args (tuple/dict):      Optional arguments to add to command,
                                particularly useful for functions.
        kwargs (dict):          Optional keyword arguments to pass to the
                                command, only used for functions.
        name (str):             Optional name of the job. If not defined,
                                guessed. If a job of the same name is
                                already queued, an integer job number (not
                                the queue number) will be added, ie.
                                <name>.1
        qtype (str):            Override the default queue type
        profile (str):          The name of a profile saved in the
                                conf
        outfile (str):          The path to the expected output file.
        outheader (bool):       Does the input outfile have a header?
        merge_func (function):  An optional function used to merge the output
                                list if there is no outfile.
        direct (bool):          Whether to run the function directly or to
                                return a Job. Default True.

        *All other keywords are parsed into cluster keywords by the
        options system. For available keywords see `fyrd.option_help()`*

    Returns:
        Primary return value varies and is decided in this order:

            If outfile:    the absolute path to that file
            If merge_func: the result of merge_func(list), where list
                           is the list of outputs.
            Else:          a list of results

        If direct is False, this function returns a fyrd.job.Job object which
        will return the results described above on get().
    """
    kwds = _options.check_arguments(kwds)
    if direct:
        return _splitrun(
            jobs, infile, inheader, command, args=args, kwargs=kwargs,
            name=name, qtype=qtype, profile=profile, outfile=outfile,
            outheader=outheader, merge_func=merge_func, **kwds
        )
    else:
        kk = dict(
            args=args, kwargs=kwargs, name=name, qtype=qtype, profile=profile,
            outfile=outfile, outheader=outheader, merge_func=merge_func
        )
        kwds = _options.sanitize_arguments(kwds)
        kk.update(_options.check_arguments(kwds))
        if callable(command):
            kk['imports']  = _run.get_all_imports(command, kk)
            kk['syspaths'] = _run.update_syspaths(command, kk)
        return _wrap_runner(
            _splitrun,
            *(jobs, infile, inheader, command),
            **kk
        )


def _splitrun(jobs, infile, inheader, command, args=None, kwargs=None,
              name=None, qtype=None, profile=None, outfile=None,
              outheader=False, merge_func=None, **kwds):
    """This is the direct running function for `splitrun()`.

    Please see that function's docstring for information and do not call this
    function directly.
    """
    # Handle arguments
    if not isinstance(jobs, int):
        raise ValueError('Jobs argument must be an integer.')
    if profile is not None and not isinstance(profile, str):
        raise ValueError('Profile must be a string, is {}'
                         .format(type(profile)))
    kwds = _options.check_arguments(kwds)

    # Get name
    name = name if name else 'split_file'

    # Check file
    infile = _os.path.abspath(_os.path.expandvars(_os.path.expanduser(infile)))
    if not _os.path.isfile(infile):
        raise OSError('Cannot find file {}'.format(infile))

    # Set up auto-cleaning
    if 'clean_files' not in kwds:
        kwds['clean_files'] = True
    if 'clean_outputs' not in kwds:
        kwds['clean_outputs'] = True

    # Get script path, returns: kwds, runpath, outpath, scriptpath
    cpath = _conf.get_job_paths(kwds)[3]

    # Prep function
    if callable(command):
        # Add all imports from the function file to globals
        kwds['imports'] = _run.export_imports(command, kwds)

    # Split file
    _logme.log('Splitting file', 'debug')
    files = _run.split_file(infile, jobs, outpath=cpath, keep_header=inheader)
    assert len(files) == jobs

    # Run the functions
    _logme.log('Submitting jobs', 'debug')
    outs = []
    count = 1
    for f in files:
        if outfile:
            o = f + '.out'
        nm = '{}_{}_of_{}'.format(name, count, jobs)
        if callable(command):
            try:
                runargs, runkwargs = _run.replace_argument(
                    [args, kwargs], '{file}', f)
            except ValueError:
                raise ValueError("No '{file}' argument in either args or " +
                                 "kwargs")
            if outfile:
                runargs, runkwargs = _run.replace_argument(
                    [runargs, runkwargs], '{outfile}', o)
            outs.append(
                _Job(command, args=runargs, kwargs=runkwargs, name=nm,
                     qtype=qtype, profile=profile, **kwds).submit()
            )
        else:
            if outfile:
                cmnd = command.format(file=f, outfile=o)
            else:
                cmnd = command.format(file=f)
            outs.append(
                _Job(cmnd, name=nm, qtype=qtype, profile=profile,
                     **kwds).submit()
            )
        count += 1

    # Get the results
    _logme.log('Waiting for results', 'debug')
    results = []
    for out in outs:
        try:
            results.append(out.get())
        except IOError:
            _logme.log('Result getting failed, most likely one of the child ' +
                       'jobs crashed, check the error files', 'critical')
            raise

    # Delete intermediate files
    _logme.log('Removing intermediate files', 'debug')
    for f in files:
        assert _os.path.isfile(f)
        _os.remove(f)

    # Return the recombined DataFrame
    _logme.log('Done, joining', 'debug')
    if outfile:
        with open(outfile, 'w') as fout:
            if outheader:
                with open(files[0]) as fin:
                    fout.write(fin.readline())
            for f in files:
                with open(f + '.out') as fin:
                    if outheader:
                        fin.readline()
                    fout.write(fin.read())
        out = _os.path.abspath(outfile)

    elif merge_func:
        assert callable(merge_func)
        out = merge_func(results)

    elif isinstance(results[0], list):
        out = _run.merge_lists(results)

    else:
        out = results

    return out


###############################################################################
#                              Helper Functions                               #
###############################################################################


def _wrap_runner(command, *args, **kkk):
    """Run a command as a job, for use with the functions in this file.

    *Note*: Only uses the 'profile', 'imports', 'syspaths', 'mem', 'clean_files',
    'clean_outputs', and 'partition' keyword arguments for job submission,
    all others are ignored.

    Returns:
        fyrd.job.Job: A Job class (not submitted) for the command.
    """
    mem = kkk['mem'] if 'mem' in kkk else 2000
    cln_f = kkk['clean_files'] if 'clean_files' in kkk else True
    cln_o = kkk['clean_outputs'] if 'clean_outputs' in kkk else True
    kkk['clean_files'] = cln_f
    kkk['clean_outputs'] = cln_o
    add_args = dict(
        cores=1, mem=mem, clean_files=cln_f, clean_outputs=cln_o
    )
    if 'partition' in kkk:
        add_args.update(dict(partition=kkk['partition']))
    if 'profile' in kkk:
        add_args.update(dict(profile=kkk['profile']))
    if callable(command):
        _logme.log('Running function {} with export run'
                   .format(command.__name__), 'debug')
        kkk['imports']  = _run.export_imports(command, kkk)
        spth = [_run.get_function_path(command)]
        kkk['syspaths'] = spth + _run.listify(kkk['syspaths']) \
            if 'syspaths' in kkk else spth
        add_args.update(dict(imports=kkk['imports'], syspaths=kkk['syspaths']))
        _logme.log('export_run imports: {}'.format(kkk['imports']), 'debug')
        _logme.log('export_run syspath: {}'.format(kkk['syspaths']), 'debug')
    else:
        _logme.log('Running command {} with export run'
                   .format(command), 'debug')
    return _Job(
        _run.export_run, (command, args, kkk), **add_args
    )



# -*- coding: utf-8 -*-
"""
Manage job dependency tracking with multiprocessing.

Runs jobs with a multiprocessing.Pool, but manages dependency using an
additional Process that loops through all submitted jobs and checks
dependencies before running.

The JobQueue class works as the queue and functions in a similar, but much more
basic, way as torque or slurm. It manages jobs by forking an instance of the
job_runner function and keeping it alive. The multiprocessing.Queue class is
then used to pass job information contained in the Job class back and forth
between the JobQueue class running in the main process and the job_runner()
fork running as a separate thread.

The actual job management is done by job_runner() and uses
multiprocessing.Process objects and not the Pool object.  This allows for more
careful management and it also allows exit codes to be captured.
"""
import os
import sys
import atexit
import signal
import multiprocessing as mp
from multiprocessing import cpu_count as _cnt
from subprocess import check_output, CalledProcessError
from time import sleep

from . import run

# Get defaults
from . import conf

# Get an Exception object to use
from . import ClusterError

# Get logging function
from . import logme

# A global placeholder for a single JobQueue instance
JQUEUE = None

__all__ = ['JobQueue']

################################
#  Normal Mode Multithreading  #
################################

THREADS  = _cnt()

# Reset broken multithreading
# Some of the numpy C libraries can break multithreading, this command
# fixes the issue.
try:
    check_output("taskset -p 0xff %d >/dev/null 2>/dev/null" % os.getpid(),
                 shell=True)
except CalledProcessError:
    pass  # This doesn't work on Macs or Windows


###############################################################################
#                      The JobQueue Class to Manage Jobs                      #
###############################################################################


class JobQueue(object):

    """Monitor and submit multiprocessing.Pool jobs with dependencies."""

    def __init__(self, cores=None):
        """Spawn a job_runner process to interact with."""
        self._jobqueue = mp.Queue()
        self._outputs  = mp.Queue()
        self.jobno     = int(conf.get_option('jobqueue', 'jobno', '1'))
        self.cores     = int(cores) if cores else THREADS
        self.runner    = mp.Process(target=job_runner,
                                    args=(self._jobqueue,
                                          self._outputs,
                                          self.cores,
                                          self.jobnumber),
                                    name='Runner')
        self.runner.start()
        self.pid = self.runner.pid
        assert self.runner.is_alive()
        self.jobs = {}

        def terminate():
            """Kill the queue runner."""
            try:
                self.runner.terminate()
                self._jobqueue.close()
                self._outputs.close()
            except AttributeError:
                pass
            if run.check_pid(self.runner.pid):
                os.kill(self.runner.pid, signal.SIGKILL)

        # Call terminate when we exit
        atexit.register(terminate)

    def update(self):
        """Get fresh job info from the runner."""
        sleep(0.5)  # This allows the queue time to flush
        if self.runner.is_alive() is not True:
            self.restart(True)
        if self.runner.is_alive() is not True:
            raise ClusterError('JobRunner has crashed')
        while not self._outputs.empty():
            # We loop through the whole queue stack, updating the dictionary
            # every time so that we get the latest info
            self.jobs.update(self._outputs.get_nowait())
        if self.jobs:
            self.jobno = max(self.jobs.keys())
            conf.set_option('jobqueue', 'jobno', str(self.jobno))

    def add(self, function, args=None, kwargs=None, dependencies=None,
            cores=1):
        """Add function to local job queue.

        Args:
            function:     A function object. To run a command, use the run.cmd
                          function here.
            args:         A tuple of args to submit to the function.
            kwargs:       A dict of keyword arguments to submit to the function.
            dependencies: A list of job IDs that this job will depend on.
            cores:        The number of threads required by this job.

        Returns:
            int: A job ID
        """
        self.update()
        assert self.runner.is_alive()
        oldjob = self.jobno
        cores = int(cores)
        if cores > self.cores:
            logme.log('Job core request exceeds resources, limiting to max: ' +
                      '{}'.format(self.cores), 'warn')
            cores = self.cores
        self._jobqueue.put(Job(function, args, kwargs, dependencies,
                               cores))
        sleep(0.5)
        self.update()
        newjob = self.jobno
        # Sometimes the queue can freeze for reasons I don't understand, this
        # is an attempted workaround.
        if not newjob == oldjob + 1:
            self.restart(True)
            self._jobqueue.put(Job(function, args, kwargs, dependencies,
                                   cores))
            self.update()
            newjob = self.jobno
        if not newjob == oldjob + 1:
            raise ClusterError('Job numbers are not updating correctly, the '
                               'local queue has probably crashed. Please '
                               'report this issue.')
        return self.jobno

    def wait(self, jobs=None):
        """Wait for a list of jobs, all jobs are the default."""
        self.update()
        if not isinstance(jobs, (list, tuple)):
            jobs = [jobs]
        while jobs:
            for job in jobs:
                if job not in self.jobs:
                    raise ClusterError('Job {} has not been submitted.'.format(job))
                if self.jobs[job].state == 'done':
                    jobs.remove(job)
                sleep(0.5)

    def get(self, job):
        """Return the output of a single job"""
        if job not in self.jobs:
            raise ClusterError('Job {} has not been submitted.'.format(job))
        self.wait(job)
        return self.jobs[job].out

    def restart(self, force=False):
        """Kill the job queue and restart it."""
        if not force:
            self.update()
            if len(self.done) != len(self.jobs):
                logme.log('Cannot restart, incomplete jobs', 'error')
                return
        self.runner.terminate()
        self.runner  = mp.Process(target=job_runner,
                                  args=(self._jobqueue, self._outputs,
                                        self.cores, self.jobnumber),
                                  name='Runner')
        self.runner.start()
        self.pid = self.runner.pid
        assert self.runner.is_alive()

    def __getattr__(self, attr):
        """Dynamic dictionary filtering."""
        if attr == 'done' or attr == 'queued' or attr == 'waiting' \
                or attr == 'running':
            newdict = {}
            for jobid, job_info in self.jobs.items():
                if job_info.state == attr:
                    newdict[jobid] = job_info
            return newdict

    def __getitem__(self, key):
        """Allow direct accessing of jobs by job id."""
        self.update()
        key = int(key)
        try:
            return self.jobs[key]
        except KeyError:
            return None

    def __iter__(self):
        """Allow us to be iterable"""
        self.update()
        for jobno, job in self.jobs.items():
            yield jobno, job

    def __len__(self):
        """Length is the total job count."""
        self.update()
        return len(self.jobs)

    def __repr__(self):
        """Class information."""
        self.update()
        return ("JobQueue<({})jobs:{};completed:{};running:{};queued:{}>"
                .format(self.cores, len(self.jobs), len(self.done),
                        len(self.running),
                        len(self.waiting) + len(self.queued)))

    def __str__(self):
        """Print jobs."""
        self.update()
        return str(self.jobs)

class Job(object):

    """An object to pass arguments to the runner."""

    def __init__(self, function, args=None, kwargs=None, depends=None,
                 cores=1):
        """Parse and save arguments."""
        if args and not isinstance(args, tuple):
            args = (args,)
        if kwargs and not isinstance(kwargs, dict):
            raise TypeError('kwargs must be a dict')
        if depends:
            if not isinstance(depends, (tuple, list)):
                depends = [depends]
            try:
                depends = [int(i) for i in depends]
            except ValueError:
                raise ValueError('dependencies must be integer job ids')
        self.function = function
        self.args     = args
        self.kwargs   = kwargs
        self.depends  = depends
        self.cores    = int(cores)

        # Assigned later
        self.id       = None
        self.pid      = None
        self.exitcode = None
        self.out      = None
        self.state    = 'Not Submitted'

    def __repr__(self):
        """Job Info."""
        return ("Job<{} (function:{},args:{}," +
                "kwargs:{};cores:{}) {}>").format(
                    self.id, self.function.__name__, self.args,
                    self.kwargs, self.cores, self.state)

    def __str__(self):
        """Print Info and Output."""
        outstr  = "Job #{}; Cores: {}\n".format(
            self.id if self.id else 'NA', self.cores)
        outstr += "\tFunction: {}\n\targs: {}\n\tkwargs: {}\n\t".format(
            self.function.__name__, self.args, self.kwargs)
        outstr += "State: {}\n\tOutput: {}\n".format(self.state, self.out)
        return outstr


###############################################################################
#               The Job Runner that will fork and run all jobs                #
###############################################################################


def job_runner(jobqueue, outputs, cores=None, jobno=None):
    """Run jobs with dependency tracking.

    Must be run as a separate multiprocessing.Process to function correctly.

    Args:
        jobqueue: A multiprocessing.Queue object into which Job objects must be
                  added. The function continually searches this Queue for new
                  jobs. Note, function must be a function call, it cannot be
                  anything else.  function is the only required argument, the
                  rest are optional.  tuples are required.
        outputs:  A multiprocessing.Queue object that will take outputs. A
                  dictionary of job objects will be output here with the
                  format:: {job_no => Job}
                  **NOTE**: function return must be picklable otherwise this
                  will raise an exception when it is put into the Queue object.
        cores:    Number of cores to use in the multiprocessing pool. Defaults
                  to all.
        jobno:    What number to start counting jobs from, default 1.
    """

    def output(out):
        """Explicitly clear the dictionary before sending the output."""
        # Don't send output if it is the same as last time.
        lastout = outputs.get() if not outputs.empty() else ''
        if out == lastout:
            return
        while not outputs.empty():
            # Clear the output object
            outputs.get()
        outputs.put(out)

    # Make sure we have Queue objects
    if not isinstance(jobqueue, mp.queues.Queue) \
            or not isinstance(outputs, mp.queues.Queue):
        raise ClusterError('jobqueue and outputs must be multiprocessing ' +
                           'Queue objects')

    # Initialize job objects
    jobno   = int(jobno) if jobno \
              else int(conf.get_option('jobqueue', 'jobno', str(1)))
    jobs    = {} # This will hold job numbers
    started = [] # A list of started jobs to check against
    cores   = cores if cores else THREADS
    queue   = [] # This will hold Processes that haven't started yet
    running = [] # This will hold actively running jobs to manage core count
    done    = [] # A list of completed jobs to check against

    # Actually loop through the jobs
    while True:
        if not jobqueue.empty():
            oldjob = jobno
            jobno += 1
            newjob = jobno
            # Sometimes the thread stalls if it has been left a while and
            # ends up reusing the same job number. I don't know why this
            # happens, however explicitly getting the jobno seems to fix the
            # issue. Just to be sure I also want to test that the job number is
            # incremented, but this is a little redundant at this point.
            assert newjob == oldjob + 1
            job = jobqueue.get_nowait()
            if not isinstance(job, Job):
                logme.log('job information must be a job object, was {}'.format(
                    type(job)), 'error')
                continue

            # The arguments look good, so lets add this to the stack.
            job.state   = 'submitted'
            job.id      = jobno
            jobs[jobno] = job

            # Send the job dictionary
            output(jobs)

        # If there are jobs, try and run them
        if jobs:
            for jobno, job_info in jobs.items():
                # Skip completed jobs
                if job_info.state == 'done':
                    continue

                # Check dependencies
                ready = True
                if job_info.depends:
                    for depend in job_info.depends:
                        if int(depend) not in done:
                            ready = False
                            job_info.state = 'waiting'
                            output(jobs)

                # Start jobs if dependencies are met and they aren't started.
                # We use daemon mode so that child jobs are killed on exit.
                if ready and not jobno in started:
                    ver = sys.version_info.major
                    # Python 2 doesn't support daemon, even though the docs
                    # say that it does.
                    gen_args = dict(name=str(jobno)) if ver == 2 \
                        else dict(name=str(jobno), daemon=True)
                    if job_info.args and job_info.kwargs:
                        queue.append((mp.Process(target=job_info.function,
                                                 args=job_info.args,
                                                 kwargs=job_info.kwargs,
                                                 **gen_args),
                                      job_info.cores))
                    elif job_info.args:
                        queue.append((mp.Process(target=job_info.function,
                                                 args=job_info.args,
                                                 **gen_args),
                                      job_info.cores))
                    elif job_info.kwargs:
                        queue.append((mp.Process(target=job_info.function,
                                                 kwargs=job_info.kwargs,
                                                 **gen_args),
                                      job_info.cores))
                    else:
                        queue.append((mp.Process(target=job_info.function,
                                                 **gen_args),
                                      job_info.cores))
                    job_info.state = 'queued'
                    started.append(jobno)
                    output(jobs)

        # Actually run jobs
        if queue:
            # Get currently used cores, ignore ourself
            running_cores = 0
            for i in [i[1] for i in running]:
                running_cores += i
            # Look for a job to run
            for j in queue:
                if j[1] + running_cores <= cores:
                    j[0].start()
                    jobs[int(j[0].name)].state = 'running'
                    jobs[int(j[0].name)].pid = j[0].pid
                    running.append(queue.pop(queue.index(j)))
                    output(jobs)
                    sleep(0.5)  # Wait for a second to allow job to start
                    break

        # Clean out running jobs
        if running:
            for i in running:
                j = i[0]
                if not j.is_alive():
                    jobs[int(j.name)].out = j.join()
                    jobs[int(j.name)].state = 'done'
                    jobs[int(j.name)].exitcode = j.exitcode
                    done.append(int(j.name))
                    running.pop(running.index(i))
                    output(jobs)

        # Wait for half a second before looping again
        sleep(0.5)



#job -*- coding: utf-8 -*-
"""
Monitor the queue for torque or slurm.

Provides a class to monitor the torque, slurm, or local queues with identical
syntax.

At its simplest, you can use it like::

    q = queue.Queue()
    q.jobs
    q.running
    q.queued
    q.complete

All of the above commands return a dictionary of:
    {job_no: Queue.QueueJob}

Queue.QueueJob classes include information on job state, owner, queue, nodes,
threads, exitcode, etc.

Queue also defines a wait() method that takes a list of job numbers, job.Job()
objects, or JobQueue.Job objects and blocks until those jobs to complete

The default cluster environment is also defined in this file as MODE, it can be
set directly or with the get_cluster_environment() function definied here.
"""
import re
import sys
import pwd      # Used to get usernames for queue
import socket   # Used to get the hostname
import getpass  # Used to get usernames for queue
from datetime import datetime as _dt
from time import time, sleep
from subprocess import check_output, CalledProcessError

# For parsing torque queues
import xml.etree.ElementTree as ET

###############################################################################
#                                Our functions                                #
###############################################################################

from . import run
from . import logme
from . import conf
from . import ClusterError

#########################
#  Which system to use  #
#########################

from . import ALLOWED_MODES

#########################################################
#  The multiprocessing pool, only used in 'local' mode  #
#########################################################

from . import local

# Funtions to import if requested
__all__ = ['Queue', 'wait', 'check_queue', 'get_cluster_environment']

# We only need the queue defaults
_defaults = conf.get_option('queue')

# This is set in the get_cluster_environment() function.
MODE = ''

# Define torque-to-slurm mappings
TORQUE_SLURM_STATES = {
    'C': 'completed',
    'E': 'completing',
    'H': 'held',  # Not a SLURM state
    'Q': 'pending',
    'R': 'running',
    'T': 'suspended',
    'W': 'running',
    'S': 'suspended',
}

# Define job states
GOOD_STATES      = ['complete', 'completed', 'special_exit']
ACTIVE_STATES    = ['configuring', 'completing', 'pending',
                    'running']
BAD_STATES       = ['boot_fail', 'cancelled', 'failed',
                    'node_fail', 'timeout', 'disappeared']
UNCERTAIN_STATES = ['hold', 'preempted', 'stopped',
                    'suspended']
ALL_STATES = GOOD_STATES + ACTIVE_STATES + BAD_STATES + UNCERTAIN_STATES
DONE_STATES = GOOD_STATES + BAD_STATES

###############################################################################
#                               The Queue Class                               #
###############################################################################


class Queue(object):

    """A wrapper for torque, slurm, or local queues.

    Attributes:
        jobs (dict)             A dictionary of all jobs in this queue
                                {jobid: Queue.QueueJob}
        max_jobs (int):         The maximum number of jobs allowed in the queue
        job_states (int):       A list of the different states of jobs in this
                                queue
        active_job_count (int): A count of all jobs that are either queued or
                                running in the current queue
        can_submit (bool):      True if total active jobs is less than max_jobs
    """

    def __init__(self, user=None, partition=None, qtype=None,):
        """Can filter by user, queue type or partition on initialization.

        Args:
            user (str):      Optional usernameto filter the queue with.
                             If user='self' or 'current', the current user will
                             be used.
            partition (str): Optional partition to filter the queue with.
            qtype (str):     'torque', 'slurm', or 'local', defaults to auto-detect.
        """
        # Get user ID as an int UID
        if user:
            if user == 'self' or user == 'current':
                self.user = getpass.getuser()
                """The username if defined."""
                self.uid  = pwd.getpwnam(self.user).pw_uid
            elif user == 'ALL':
                self.user = None
            else:
                if isinstance(user, int) \
                        or (isinstance(user, str) and user.isdigit()):
                    self.uid  = int(user)
                else:
                    self.uid = pwd.getpwnam(str(user)).pw_uid
        else:
            self.uid = None
        self.user = pwd.getpwuid(self.uid).pw_name if self.uid else None
        self.partition = partition
        """The partition if defined."""

        # Set queue length
        self.max_jobs = conf.get_option('queue', 'max_jobs')
        """The maximum number of jobs that can run in this queue."""

        # Support python2, which hates reciprocal import
        from .job import Job
        from .local import Job as QJob
        self._Job      = Job
        self._JobQueue = QJob

        # Get sleep time and update time
        self.queue_update_time = conf.get_option('queue', 'queue_update', 2)
        self.sleep_len         = conf.get_option('queue', 'sleep_len', 2)

        # Set type
        if qtype:
            check_queue(qtype)
        else:
            check_queue()
        self.qtype = qtype if qtype else MODE

        # Allow tracking of updates to prevent too many updates
        self._updating = False

        # Will contain a dict of QueueJob objects indexed by ID
        self.jobs = {}
        """All jobs currently in this queue."""

        self._update()

    ####################
    #  Public Methods  #
    ####################

    def check_dependencies(self, dependencies):
        """Check if dependencies are running.

        Args:
            dependencies (list): List of job IDs

        Returns:
            str: 'active' if dependencies are running or queued, 'good' if
                 completed, 'bad' if failed, cancelled, or suspended, 'absent'
                 otherwise.
        """
        for dep in run.listify(dependencies):
            dep = int(dep)
            if dep not in self.jobs:
                return 'absent'
            state = self.jobs[dep].state
            if state in ACTIVE_STATES:
                return 'active'
            elif state in GOOD_STATES:
                return 'good'
            elif state in BAD_STATES or state in UNCERTAIN_STATES:
                return 'bad'

    def wait(self, jobs):
        """Block until all jobs in jobs are complete.

        Update time is dependant upon the queue_update parameter in
        your ~/.fyrd file.

        In addition, wait() will not return until between 1 and 3
        seconds after a job has completed, irrespective of queue_update
        time. This allows time for any copy operations to complete after
        the job exits.

        Args:
            jobs:  A job or list of jobs to check. Can be one of: Job or
                   multiprocessing.pool.ApplyResult objects, job ID (int/str),
                   or a object or a list/tuple of multiple Jobs or job IDs.

        Returns:
            True on success False or None on failure.
        """
        self.update()
        logme.log('Queue waiting.', 'debug')

        # Sanitize arguments
        if not isinstance(jobs, (list, tuple)):
            jobs = [jobs]
        for job in jobs:
            if not isinstance(job, (str, int, self.QueueJob, self._Job,
                                    self._JobQueue)):
                raise ClusterError('job must be int, string, or Job, ' +
                                   'is {}'.format(type(job)))

        # Wait for 0.1 second before checking, as jobs take a while to be
        # queued sometimes
        sleep(0.1)
        for job in jobs:
            logme.log('Checking {}'.format(job), 'debug')
            qtype = job.qtype if isinstance(job, self._Job) else self.qtype
            if isinstance(job, (self._Job, self._JobQueue, self.QueueJob)):
                job = job.id
            if qtype == 'local':
                logme.log('Job is in local queue', 'debug')
                try:
                    job = int(job)
                except TypeError:
                    raise TypeError('Job must be a Job object or job #.')
                if not local.JQUEUE \
                        or not local.JQUEUE.runner.is_alive():
                    raise ClusterError('Cannot wait on job ' + str(job) +
                                       'JobQueue does not exist')
                local.JQUEUE.wait(job)
            else:
                logme.log('Job is in remote queue', 'debug')
                if isinstance(job, self._Job):
                    job = job.id
                job = int(job)
                not_found = 0
                lgd = False
                while True:
                    self._update()
                    # Allow 12 seconds to elapse before job is found in queue,
                    # if it is not in the queue by then, assume completion.
                    if job not in self.jobs:
                        if lgd:
                            logme.log('Attempt #{}/12'.format(not_found),
                                      'debug')
                        else:
                            logme.log('{} not in queue, waiting up to 12s '
                                      .format(job) +
                                      'for it to appear', 'info')
                            lgd = True
                        sleep(1)
                        not_found += 1
                        if not_found == 12:
                            logme.log(
                                '{} not in queue, tried 12 times over 12s'
                                .format(job) + '. Job likely completed, ' +
                                'assuming completion, stats will be ' +
                                'unavailable.','warn'
                            )
                            return 'disappeared'
                        continue
                    ## Actually look for job in running/queued queues
                    lgd      = False
                    lgd2     = False
                    start    = _dt.now()
                    res_time = conf.get_option('queue', 'res_time')
                    count    = 0
                    # Get job state
                    job_state = self.jobs[job].state
                    # Check the state
                    if job_state in GOOD_STATES:
                        logme.log('Queue wait for {} complete'
                                  .format(job), 'debug')
                        sleep(0.1)
                        break
                    elif job_state in ACTIVE_STATES:
                        if lgd:
                            logme.log('{} not complete yet, waiting'
                                      .format(job), 'debug')
                            lgd = True
                        else:
                            logme.log('{} still not complete, waiting'
                                      .format(job), 'verbose')
                        sleep(self.sleep_len)
                    elif job_state in BAD_STATES:
                        logme.log('Job {} failed with state {}'
                                  .format(job, job_state), 'error')
                        return False
                    elif job_state in UNCERTAIN_STATES:
                        if not lgd2:
                            logme.log('Job {} in state {}, waiting {} '
                                      .format(job, job_state, res_time) +
                                      'seconds for resolution', 'warn')
                            lgd2 = True
                        if (_dt.now() - start).seconds > res_time:
                            logme.log('Job {} still in state {}, aborting'
                                      .format(job, job_state), 'error')
                            return False
                        sleep(self.sleep_len)
                    else:
                        if count == 5:
                            logme.log('Job {} in unknown state {} '
                                      .format(job, job_state) +
                                      'cannot continue', 'critical')
                            raise QueueError('Unknown job state {}'
                                             .format(job_state))
                        logme.log('Job {} in unknown state {} '
                                  .format(job, job_state) +
                                  'trying to resolve', 'debug')
                        count += 1
                        sleep(self.sleep_len)

        # Sleep an extra half second to allow post-run scripts to run
        sleep(0.5)
        return True

    def wait_to_submit(self, max_jobs=None):
        """Block until fewer running/queued jobs in queue than max_jobs.

        Args:
            max_jobs (int): Override self.max_jobs
        """
        count   = 50
        written = False
        if max_jobs:
            self.max_jobs = int(max_jobs)
        while True:
            if self.can_submit:
                return
            if not written:
                logme.log(('The queue is full, there are {} jobs running and '
                           '{} jobs queued. Will wait to submit, retrying '
                           'every {} seconds.')
                          .format(len(self.running), len(self.queued),
                                  self.sleep_len),
                          'info')
                written = True
            if count == 0:
                logme.log('Still waiting to submit.', 'info')
                count = 50
            count -= 1
            sleep(self.sleep_len)

    def update(self):
        """Refresh the list of jobs from the server, limit queries."""
        if int(time()) - self.last_update > self.queue_update_time:
            self._update()
        else:
            logme.log('Skipping update as last update too recent', 'debug')
        return self

    def get_jobs(self, key):
        """Return a dict of jobs where state matches key."""
        retjobs = {}
        for jobid, job in self.jobs.items():
            if job.state == key.lower():
                retjobs[jobid] = job
        return retjobs

    def get_user_jobs(self, users):
        """Filter jobs by user.

        Args:
            users (list): A list of users/owners

        Returns:
            dict: A filtered job dictionary of {job_id: QueueJob} for all jobs
                  owned by the queried users.
        """
        try:
            if isinstance(users, (str, int)):
                users = [users]
            else:
                users = list(users)
        except TypeError:
            users = [users]
        return {k: v for k, v in self.jobs.items() if v.owner in users}

    @property
    def users(self):
        """Return a list of users with jobs running."""
        return [job.owner for job in self.jobs.values()]

    @property
    def job_states(self):
        """Return a list of job states for all jobs in the queue."""
        return [job.state for job in self.jobs]

    @property
    def finished(self):
        """Return a list of jobs that are neither queued nor running."""
        return {i: j for i, j in self.jobs.items() \
                if j.state not in ACTIVE_STATES}

    @property
    def bad(self):
        """Return a list of jobs that have bad or uncertain states."""
        return {i: j for i, j in self.jobs.items() \
                if j.state in BAD_STATES or j.state in UNCERTAIN_STATES}

    @property
    def active_job_count(self):
        """Return a count of all queued or running jobs."""
        return sum([
            len(j) for j in [
                self.get_jobs(i) for i in ACTIVE_STATES
            ]
        ])

    @property
    def can_submit(self):
        """Return True if R/Q jobs are less than max_jobs.

        If max_jobs is None, default from config is used.
        """
        self.update()
        return self.active_job_count < self.max_jobs

    ######################
    # Internal Functions #
    ######################

    def _update(self):
        """Refresh the list of jobs from the server.

        This is the core queue interaction function of this class.
        """
        if self._updating:
            return
        logme.log('Queue updating', 'debug')
        # Set the update time I don't care about microseconds
        self.last_update = int(time())

        jobs = []  # list of jobs created this session

        # Mode specific initialization
        if self.qtype == 'local':
            if not local.JQUEUE or not local.JQUEUE.runner.is_alive():
                local.JQUEUE = local.JobQueue(cores=local.THREADS)
            for job_id, job_info in local.JQUEUE:
                if job_id in self.jobs:
                    job = self.jobs[job_id]
                else:
                    job = self.QueueJob()
                job.id     = job_id
                job.name   = job_info.function.__name__
                job.owner  = self.user
                self.nodes = socket.gethostname()
                if job_info.state == 'Not Submitted':
                    job.state = 'pending'
                elif job_info.state == 'waiting' \
                        or job_info.state == 'submitted':
                    job.state = 'pending'
                elif job_info.state == 'started' \
                        or job_info.state == 'running':
                    job.state = 'running'
                elif job_info.state == 'done':
                    job.state = 'completed'
                    job.exitcode = int(job_info.exitcode)
                elif job_info.state == 'queued':
                    job.state = 'pending'
                else:
                    raise Exception('Unrecognized state')

                # Assign the job to self.
                self.jobs[job_id] = job
                jobs.append(job_id)

        else:
            for [job_id, job_name, job_user, job_partition,
                 job_state, job_nodelist, job_nodecount,
                 job_cpus, job_exitcode] in queue_parser(self.qtype,
                                                         self.user,
                                                         self.partition):
                if job_id not in self.jobs:
                    job = self.QueueJob()
                else:
                    job = self.jobs[job_id]
                jobs.append(job_id)
                job.id    = job_id
                job.name  = job_name
                job.owner = job_user
                job.queue = job_partition
                job.state = job_state.lower()
                job.nodes = job_nodelist

                # Threads is number of nodes * jobs per node
                if job_nodecount and job_cpus:
                    job.threads = int(job_nodecount) * int(job_cpus)
                else:
                    job.threads = None
                if job.state == 'completed' or job.state == 'failed':
                    job.exitcode = job_exitcode

                # Assign the job to self.
                self.jobs[job_id] = job
                jobs.append(job_id)

        # We assume that if a job just disappeared it completed
        if self.jobs:
            for qjob in self.jobs.values():
                if qjob.id not in jobs:
                    qjob.state = 'completed'
                    qjob.disappeared = True

    def __getattr__(self, key):
        """Make running and queued attributes dynamic."""
        key = key.lower()
        if key in TORQUE_SLURM_STATES:
            key = TORQUE_SLURM_STATES[key]
        if key == 'complete':
            key = 'completed'
        elif key == 'queued':
            key = 'pending'
        if key in ALL_STATES:
            return self.get_jobs(key)

    def __getitem__(self, key):
        """Allow direct accessing of jobs by job id."""
        if isinstance(key, self._Job):
            key = key.jobid
        key = int(key)
        try:
            return self.jobs[key]
        except KeyError:
            return None

    def __iter__(self):
        """Allow us to be iterable."""
        for jb in self.jobs.values():
            yield jb

    def __len__(self):
        """Length is the total job count."""
        return len(self.jobs)

    def __repr__(self):
        """For debugging."""
        self._updating = True
        if self.user:
            outstr = 'Queue<jobs:{};completed:{};queued:{};user={}>'.format(
                len(self), len(self.complete), len(self.queued), self.user)
        else:
            outstr = 'Queue<jobs:{};completed:{};queued:{};user=ALL>'.format(
                len(self), len(self.complete), len(self.queued))
        self._updating = False
        return outstr

    def __str__(self):
        """A list of keys."""
        return str(self.jobs.keys())

    ##############################################
    #  A simple class to hold jobs in the queue  #
    ##############################################

    class QueueJob(object):

        """A very simple class to store info about jobs in the queue.

        Only used for torque and slurm queues.

        Attributes:
            id (int):           Job ID
            name (str):         Job name
            owner (str):        User who owns the job
            threads (int):      Number of cores used by the job
            queue (str):        The queue/partition the job is running in
            state (str):        Current state of the job, normalized to slurm
                                states
            nodes (list):       List of nodes job is running on
            exitcode (int):     Exit code of completed job
            disappeared (bool): Job cannot be found in the queue anymore
        """

        id          = None
        name        = None
        owner       = None
        threads     = None
        queue       = None
        state       = None
        nodes       = None
        exitcode    = None
        disappeared = False

        def __init__(self):
            """No initialization needed all attributes are set elsewhere."""
            pass

        def __repr__(self):
            """Show all info."""
            outstr = ("Queue.QueueJob<{id}:{state}({name},owner:{owner}," +
                      "queue:{queue},nodes:{nodes},threads:{threads}," +
                      "exitcode:{code})").format(
                          id=self.id, name=self.name, owner=self.owner,
                          queue=self.queue, nodes=self.nodes,
                          code=self.exitcode, threads=self.threads,
                          state=self.state)
            if self.disappeared:
                outstr += 'DISAPPEARED>'
            else:
                outstr += '>'
            return outstr

        def __str__(self):
            """Print job ID."""
            return str(self.id)

################
#  Exceptions  #
################

class QueueError(Exception):

    """Simple Exception wrapper."""

    pass


###############################################################################
#                             Non-Class Functions                             #
###############################################################################


###################
#  Queue Parsers  #
###################

def queue_parser(qtype=None, user=None, partition=None):
    """Call either torque or slurm qtype parsers depending on qtype.

    Args:
        qtype: Either 'torque' or 'slurm', defaults to current MODE
        user:  optional user name to pass to queue to filter queue with

    Yields:
        tuple: job_id, name, userid, partition, state, nodelist, numnodes,
               ntpernode, exit_code
    """
    if not qtype:
        qtype = get_cluster_environment()
    if qtype == 'torque':
        return torque_queue_parser(user, partition)
    elif qtype == 'slurm':
        return slurm_queue_parser(user, partition)
    else:
        raise ClusterError("Invalid qtype type {}, must be 'torque' or 'slurm'"
                           .format(qtype))


def torque_queue_parser(user=None, partition=None):
    """Iterator for torque queues.

    Use the `qstat -x` command to get an XML queue for compatibility.

    Args:
        user:     optional user name to pass to qstat to filter queue with
        partiton: optional partition to filter the queue with

    Yields:
        tuple: job_id, name, userid, partition, state, nodelist, numnodes,
               ntpernode, exit_code

    numcpus is currently always 1 as most torque queues treat every core as a
    node.
    """
    # I am not using run.cmd because I want to catch XML errors also
    try_count = 0
    qargs = ['qstat', '-x']
    while True:
        try:
            xmlqueue = ET.fromstring(check_output(qargs))
        except CalledProcessError:
            sleep(1)
            if try_count == 5:
                raise
            else:
                try_count += 1
        except ET.ParseError:
            # ElementTree throws error when string is empty
            sleep(1)
            if try_count == 1:
                xmlqueue = None
                break
            else:
                try_count += 1
        else:
            break

    # Create QueueJob objects for all entries that match user
    if xmlqueue is not None:
        for xmljob in xmlqueue:
            job_id = xmljob.find('Job_Id').text.split('.')[0]
            if '[' in job_id:
                job_id, array_id = job_id.split('[')
                array_id = array_id.strip('[]')
                if array_id:
                    array_id = int(array_id)
                else:
                    array_id = 0
            else:
                array_id = None
            try:
                job_id = int(job_id)
            except ValueError:
                # Allow string job IDs
                pass

            job_owner = xmljob.find('Job_Owner').text.split('@')[0]
            if user and job_owner != user:
                continue
            job_name  = xmljob.find('Job_Name').text
            job_queue = xmljob.find('queue').text
            job_state = xmljob.find('job_state').text
            job_state = TORQUE_SLURM_STATES[job_state]
            logme.log('Job {} state: {}'.format(job_id, job_state),
                      'debug')
            ndsx = xmljob.find('exec_host')
            if ndsx:
                nds = ndsx.text.split('+')
            else:
                nds = []
            nodes = []
            for node in nds:
                if '-' in node:
                    nm, num = node.split('/')
                    for i in range(*[int(i) for i in num.split('-')]):
                        nodes.append(nm + '/' + str(i).zfill(2))
                else:
                    nodes.append(node)
            # I assume that every 'node' is a core, as that is the
            # default for torque, but it isn't always true
            job_threads  = len(nodes)
            exitcode     = xmljob.find('exit_status')
            if hasattr(exitcode, 'text'):
                exitcode = int(exitcode.text)

            if partition and job_queue != partition:
                continue
            yield (job_id, job_name, job_owner, job_queue, job_state,
                   nodes, job_threads, 1, exitcode)


def slurm_queue_parser(user=None, partition=None):
    """Iterator for slurm queues.

    Use the `squeue -O` command to get standard data across implementation,
    supplement this data with the results of `sacct`. sacct returns data only
    for the current user but retains a much longer job history. Only jobs not
    returned by squeue are added with sacct, and they are added to *the end* of
    the returned queue, i.e. *out of order with respect to the actual queue*.

    Args:
        user:      optional user name to filter queue with
        partition: optional partition to filter queue with

    Yields:
        tuple: job_id, name, userid, partition, state, nodelist, numnodes,
               ntpernode, exit_code
    """
    nodequery = re.compile(r'([^\[,]+)(\[[^\[]+\])?')
    qargs = ['squeue', '-h', '-O',
             'jobid:400,arraytaskid:400,name:400,userid:400,partition:400,' +
             'state:400,nodelist:400,numnodes:400,numcpus:400,exit_code:400']
    # Parse queue info by length
    squeue = [
        tuple(
            [k[i:i+200].rstrip() for i in range(0, 4000, 400)]
        ) for k in run.cmd(qargs)[1].split('\n')
    ]
    # SLURM sometimes clears the queue extremely fast, so we use sacct
    # to get old jobs by the current user
    qargs = ['sacct', '-p',
             '--format=jobid,jobname,user,partition,state,' +
             'nodelist,reqnodes,ncpus,exitcode']
    try:
        sacct = [tuple(i.strip(' |').split('|')) for i in
                 run.cmd(qargs)[1].split('\n')]
        sacct = sacct[1:]
    # This command isn't super stable and we don't care that much, so I will
    # just let it die no matter what
    except Exception as e:
        if logme.MIN_LEVEL == 'debug':
            raise e
        else:
            sacct = []

    if sacct:
        if len(sacct[0]) != 9:
            logme.log('sacct parsing failed unexpectedly as there are not ' +
                      '9 columns, aborting.', 'critical')
            raise ValueError('sacct output does not have 9 columns. Has:' +
                             '{}: {}'.format(len(sacct[0]), sacct[0]))
        jobids = [i[0] for i in squeue]
        for sinfo in sacct:
            # Skip job steps, only index whole jobs
            if '.' in sinfo[0]:
                logme.log('Skipping {} '.format(sinfo[0]) +
                          "in sacct processing as it is a job part.",
                          'verbose')
                continue
            # These are the values I expect
            try:
                [sid, sname, suser, spartition, sstate,
                 snodelist, snodes, scpus, scode] = sinfo
                if '_' in sid:
                    sid, sarr = sid.split('_')
                    sif = '{}_{}'.format(sid, sarr)
                else:
                    sarr = None
                    sif = '{}'.format(sid)
            except ValueError as err:
                logme.log('sacct parsing failed with error {} '.format(err) +
                          'due to an incorrect number of entries.\n' +
                          'Contents of sinfo:\n{}\n'.format(sinfo) +
                          'Expected 9 values\n:' +
                          '[sid, sname, suser, spartition, sstate, ' +
                          'snodelist, snodes, scpus, scode]',
                          'critical')
                raise
            # Skip jobs that were already in squeue
            if sid in jobids:
                logme.log('{} still in squeue output'.format(sid),
                          'verbose')
                continue
            scode = int(scode.split(':')[-1])
            squeue.append((sid, sarr, sname, suser, spartition, sstate,
                           snodelist, snodes, scpus, scode))
    else:
        logme.log('No job info in sacct', 'debug')

    # Sanitize data
    for sinfo in squeue:
        if len(sinfo) == 10:
            [sid, sarr, sname, suser, spartition, sstate, sndlst,
             snodes, scpus, scode] = sinfo
        else:
            sys.stderr.write('{}'.format(repr(sinfo)))
            raise ClusterError('Queue parsing error, expected 10 items '
                               'in output of squeue and sacct, got {}\n'
                               .format(len(sinfo)))
        if partition and spartition != partition:
            continue
        if not isinstance(sid, int):
            sid = int(sid) if sid else None
        if isinstance(sarr, str) and sarr.isdigit():
            sarr = int(sarr)
        else:
            sarr = None
        if not isinstance(snodes, int):
            snodes = int(snodes) if snodes else None
        if not isinstance(scpus, int):
            scpus = int(scpus) if snodes else None
        if not isinstance(scode, int):
            scode = int(scode) if scode else None
        # Convert user from ID to name
        if suser.isdigit():
            suser = pwd.getpwuid(int(suser)).pw_name
        if user and suser != user:
            continue
        # Attempt to parse nodelist
        snodelist = []
        if sndlst:
            if nodequery.search(sndlst):
                nsplit = nodequery.findall(sndlst)
                for nrg in nsplit:
                    node, rge = nrg
                    if not rge:
                        snodelist.append(node)
                    else:
                        for reg in rge.strip('[]').split(','):
                            # Node range
                            if '-' in reg:
                                start, end = [int(i) for i in reg.split('-')]
                                for i in range(start, end):
                                    snodelist.append('{}{}'.format(node, i))
                            else:
                                snodelist.append('{}{}'.format(node, reg))
            else:
                snodelist = sndlst.split(',')

        yield (sid, sname, suser, spartition, sstate, snodelist,
               snodes, scpus, scode)


###########################################################
#  Set the global cluster type: slurm, torque, or local  #
###########################################################


def get_cluster_environment():
    """Detect the local cluster environment and set MODE globally.

    Uses which to search for sbatch first, then qsub. If neither is found,
    MODE is set to local.

    Returns:
        tuple: MODE variable ('torque', 'slurm', or 'local')
    """
    global MODE
    conf_queue = conf.get_option('queue', 'queue_type', 'auto')
    if conf_queue not in ['torque', 'slurm', 'local', 'auto']:
        logme.log('queue_type in the config file is {}, '.format(conf_queue) +
                  'but it should be one of torque, slurm, local, or auto. ' +
                  'Resetting it to auto', 'warn')
        conf.set_option('queue', 'queue_type', 'auto')
        conf_queue = 'auto'
    if conf_queue == 'auto':
        sbatch_cmnd = conf.get_option('queue', 'sbatch')
        qsub_cmnd   = conf.get_option('queue', 'qsub')
        sbatch_cmnd = sbatch_cmnd if sbatch_cmnd else 'sbatch'
        qsub_cmnd   = qsub_cmnd if qsub_cmnd else 'qsub'
        if run.which(sbatch_cmnd):
            MODE = 'slurm'
        elif run.which(qsub_cmnd):
            MODE = 'torque'
        else:
            MODE = 'local'
    else:
        MODE = conf_queue
    if MODE == 'slurm' or MODE == 'torque':
        logme.log('{} detected, using for cluster submissions'.format(MODE),
                  'debug')
    else:
        logme.log('No cluster environment detected, using multiprocessing',
                  'debug')
    return MODE


##############################
#  Check if queue is usable  #
##############################


def check_queue(qtype=None):
    """Raise exception if MODE is incorrect."""
    if 'MODE' not in globals():
        global MODE
        MODE = get_cluster_environment()
    if not MODE:
        MODE = get_cluster_environment()
    if qtype:
        if qtype not in ALLOWED_MODES:
            raise ClusterError('qtype value {} is not recognized, '
                               .format(qtype) +
                               'should be: local, torque, or slurm')
        else:
            if MODE not in ALLOWED_MODES:
                MODE = qtype
            return True
    elif MODE not in ALLOWED_MODES:
        raise ClusterError('MODE value {} is not recognized, '.format(MODE) +
                           'should be: local, torque, or slurm')


######################################################################
#  Expose the Queue waiting method without requiring a Queue object  #
######################################################################


def wait(jobs):
    """Wait for jobs to finish.

    Args:
        jobs: A single job or list of jobs to wait for. With torque or slurm,
              these should be job IDs, with local mode, these are
              multiprocessing job objects (returned by submit())
    """
    # Support python2, which hates reciprocal import for 80's reasons
    from .job import Job
    from .local import JobQueue

    check_queue()  # Make sure the MODE is usable

    # Sanitize argument
    if not isinstance(jobs, (list, tuple)):
        jobs = [jobs]
    for job in jobs:
        if not isinstance(job, (str, int, Job, JobQueue)):
            raise ClusterError('job must be int, string, or Job, ' +
                               'is {}'.format(type(job)))

    if MODE == 'local':
        for job in jobs:
            try:
                job = int(job)
            except TypeError:
                raise TypeError('Job must be a Job object or job #.')
            if not local.JQUEUE or not local.JQUEUE.runner.is_alive():
                raise ClusterError('Cannot wait on job ' + str(job) +
                                   'JobQueue does not exist')
            local.JQUEUE.wait(job)

    elif MODE == 'torque':
        # Wait for 1 seconds before checking, as jobs take a while to be queued
        # sometimes
        sleep(1)

        s = re.compile(r' +')  # For splitting qstat output
        # Jobs must be strings for comparison operations
        jobs = [str(j) for j in jobs]
        q = run.cmd('qstat -a', tries=8)[1].rstrip().split('\n')
        # Check header
        if not re.split(r' {2,100}', q[3])[9] == 'S':
            raise ClusterError('Unrecognized torque qstat format')
        # Build a list of completed jobs
        complete = []
        for j in q[5:]:
            i = s.split(j)
            if i[9] == 'C':
                complete.append(i[0].split('.')[0])
        # Build a list of all jobs
        alljobs  = [s.split(j)[0].split('.')[0] for j in q[5:]]
        # Trim down job list
        jobs = [j for j in jobs if j in alljobs]
        jobs = [j for j in jobs if j not in complete]
        if len(jobs) == 0:
            return
    elif MODE == 'slurm':
        # Wait for 2 seconds before checking, as jobs take a while to be queued
        # sometimes
        sleep(2)

        # Jobs must be strings for comparison operations
        jobs = [str(j) for j in jobs]
        while True:
            # Slurm allows us to get a custom output for faster parsing
            q = check_output(
                ['squeue', '-h', '-o', "'%A,%t'"]).decode().rstrip().split(',')
            # Build a list of jobs
            complete = [i[0] for i in q if i[1] == 'CD']
            failed   = [i[0] for i in q if i[1] == 'F']
            allj     = [i[0] for i in q]
            # Trim down job list, ignore failures
            jobs = [i for i in jobs if i not in allj]
            jobs = [i for i in jobs if i not in complete]
            jobs = [i for i in jobs if i not in failed]
            if len(jobs) == 0:
                return
            sleep(2)



# -*- coding: utf-8 -*-
"""
Class and methods to handle Job submission.
"""
import os  as _os
import sys as _sys
from uuid import uuid4 as _uuid
from time import sleep as _sleep
from datetime import datetime as _dt
from subprocess import CalledProcessError as _CalledProcessError

# Try to use dill, revert to pickle if not found
import dill as _pickle
from six import reraise as _reraise

###############################################################################
#                                Our functions                                #
###############################################################################

from . import run     as _run
from . import conf    as _conf
from . import queue   as _queue
from . import logme   as _logme
from . import local   as _local
from . import options as _options
from . import script_runners as _scrpts
from . import ClusterError   as _ClusterError
from .submission_scripts import Script   as _Script
from .submission_scripts import Function as _Function


__all__ = ['Job']

###############################################################################
#                                The Job Class                                #
###############################################################################


class Job(object):

    """Information about a single job on the cluster.

    Holds information about submit time, number of cores, the job script,
    and more.

    Below are the core attributes and methods required to use this class.

    Attributes:
        out (str):            The output of the function or a copy of stdout
                              for a script
        stdout (str):         Any output to STDOUT
        stderr (str):         Any output to STDERR
        exitcode (int):       The exitcode of the running processes (the script
                              runner if the Job is a function.
        start (datetime):     A datetime object containing time execution
                              started on the remote node.
        end (datetime):       Like start but when execution ended.
        runtime (timedelta):  A timedelta object containing runtime.
        files (list):         A list of script files associated with this class
        done (bool):          True if the job has completed

    Methods:
        submit(): submit the job if it is ready
        wait():   block until the job is done
        get():    block until the job is done and then return the output
                  (stdout if job is a script), by default saves all outputs to
                  self (i.e. .out, .stdout, .stderr) and deletes all
                  intermediate files before returning. If `save` argument is
                  `False`, does not delete the output files by default.
        clean():  delete any files created by this object

    Printing or reproducing the class will display detailed job information.

    Both `wait()` and `get()` will update the queue every few seconds
    (defined by the queue_update item in the config) and add queue information
    to the job as they go.

    If the job disappears from the queue with no information, it will be listed
    as 'completed'.

    All jobs have a .submission attribute, which is a Script object containing
    the submission script for the job and the file name, plus a 'written' bool
    that checks if the file exists.

    In addition, SLURM jobs have a .exec_script attribute, which is a Script
    object containing the shell command to _run. This difference is due to the
    fact that some SLURM systems execute multiple lines of the submission file
    at the same time.

    Finally, if the job command is a function, this object will also contain a
    `.function` attribute, which contains the script to run the function.

    """

    id            = None
    submitted     = False
    written       = False
    found         = False
    submit_time   = None

    # Holds a pool object if we are in local mode
    pool_job      = None

    # Scripts
    submission    = None
    exec_script   = None
    function      = None
    imports       = None

    # Dependencies
    dependencies  = None

    # Pickled output file for functions
    poutfile      = None

    # Holds queue information in torque and slurm
    queue_info    = None

    # Output tracking
    _got_out      = False
    _got_stdout   = False
    _got_stderr   = False
    _got_exitcode = False
    _out          = None
    _stdout       = None
    _stderr       = None
    _exitcode     = None

    # Time tracking
    _got_times    = False
    start         = None
    end           = None

    # Track update status
    _updating     = False

    # Auto Cleaning
    clean_files   = _conf.get_option('jobs', 'clean_files')
    clean_outputs = _conf.get_option('jobs', 'clean_outputs')

    def __init__(self, command, args=None, kwargs=None, name=None, qtype=None,
                 profile=None, **kwds):
        """Initialization function arguments.

        Args:
            command (function/str): The command or function to execute.
            args (tuple/dict):      Optional arguments to add to command,
                                    particularly useful for functions.
            kwargs (dict):          Optional keyword arguments to pass to the
                                    command, only used for functions.
            name (str):             Optional name of the job. If not defined,
                                    guessed. If a job of the same name is
                                    already queued, an integer job number (not
                                    the queue number) will be added, ie.
                                    <name>.1
            qtype (str):            Override the default queue type
            profile (str):          The name of a profile saved in the
                                    conf

            *All other keywords are parsed into cluster keywords by the
            options system. For available keywords see `fyrd.option_help()`*
        """

        ########################
        #  Sanitize arguments  #
        ########################
        _logme.log('Args pre-check: {}'.format(kwds), 'debug')
        kwds = _options.check_arguments(kwds)
        _logme.log('Args post-check: {}'.format(kwds), 'debug')

        # Override autoclean state (set in config file)
        if 'clean_files' in kwds:
            self.clean_files = kwds.pop('clean_files')
        if 'clean_outputs' in kwds:
            self.clean_outputs = kwds.pop('clean_outputs')

        # Path handling
        [kwds, self.runpath,
         self.outpath, self.scriptpath] = _conf.get_job_paths(kwds)

        # Save command
        self.command = command
        self.args    = args

        # Merge in profile, this includes all args from the DEFAULT profile
        # as well, ensuring that those are always set at a minumum.
        profile = profile if profile else 'DEFAULT'
        prof = _conf.get_profile(profile)
        if not prof:
            raise _ClusterError('No profile found for {}'.format(profile))
        for k,v in prof.args.items():
            if k not in kwds:
                kwds[k] = v

        # Use the default profile as a backup if any arguments missing
        default_args = _conf.DEFAULT_PROFILES['DEFAULT']
        default_args.update(_conf.get_profile('DEFAULT').args)
        for opt, arg in default_args.items():
            if opt not in kwds:
                _logme.log('{} not in kwds, adding from default: {}:{}'
                           .format(opt, opt, arg), 'debug')
                kwds[opt] = arg

        # Get environment
        if not _queue.MODE:
            _queue.MODE = _queue.get_cluster_environment()
        self.qtype = qtype if qtype else _queue.MODE
        self.queue = _queue.Queue(user='self', qtype=self.qtype)
        self.state = 'Not_Submitted'

        # Set name
        if not name:
            if callable(command):
                strcmd = str(command).strip('<>')
                parts = strcmd.split(' ')
                if parts[0] == 'bound':
                    name = '_'.join(parts[2:3])
                else:
                    parts.remove('function')
                    try:
                        parts.remove('built-in')
                    except ValueError:
                        pass
                    name = parts[0]
            else:
                name = command.split(' ')[0].split('/')[-1]

        # Make sure name not in queue
        self.uuid = str(_uuid()).split('-')[0]
        names     = [i.name.split('.')[0] for i in self.queue]
        namecnt   = len([i for i in names if i == name])
        name      = '{}.{}.{}'.format(name, namecnt, self.uuid)
        self.name = name

        # Set modules
        self.modules = kwds.pop('modules') if 'modules' in kwds else None
        if self.modules:
            self.modules = _run.opt_split(self.modules, (',', ';'))

        # Make sure args are a tuple or dictionary
        if args:
            if isinstance(args, str):
                args = tuple(args)
            if not isinstance(args, (tuple, dict)):
                try:
                    args = tuple(args)
                except TypeError:
                    args = (args,)

        # In case cores are passed as None
        if 'nodes' not in kwds:
            kwds['nodes'] = default_args['nodes']
        if 'cores' not in kwds:
            kwds['cores'] = default_args['cores']
        self.nodes = kwds['nodes']
        self.cores = kwds['cores']

        # Set output files
        suffix = kwds.pop('suffix') if 'suffix' in kwds \
                 else _conf.get_option('jobs', 'suffix')
        if 'outfile' in kwds:
            pth, fle = _os.path.split(kwds['outfile'])
            if not pth:
                pth = self.outpath
            kwds['outfile'] = _os.path.join(pth, fle)
        else:
            kwds['outfile'] = _os.path.join(
                self.outpath, '.'.join([name, suffix, 'out']))
        if 'errfile' in kwds:
            pth, fle = _os.path.split(kwds['errfile'])
            if not pth:
                pth = self.outpath
            kwds['errfile'] = _os.path.join(pth, fle)
        else:
            kwds['errfile'] = _os.path.join(
                self.outpath, '.'.join([name, suffix, 'err']))
        self.outfile = kwds['outfile']
        self.errfile = kwds['errfile']

        # Check and set dependencies
        if 'depends' in kwds:
            dependencies = _run.listify(kwds.pop('depends'))
            self.dependencies = []
            errmsg = 'Dependencies must be number or list'
            for dependency in dependencies:
                if isinstance(dependency, str):
                    if not dependency.isdigit():
                        raise _ClusterError(errmsg)
                    dependency  = int(dependency)
                if not isinstance(dependency, (int, Job)):
                    raise _ClusterError(errmsg)
                self.dependencies.append(dependency)

        ######################################
        #  Command and Function Preparation  #
        ######################################

        # Get imports
        imports = kwds.pop('imports') if 'imports' in kwds else None

        # Get syspaths
        syspaths = kwds.pop('syspaths') if 'syspaths' in kwds else None

        # Split out sys.paths from imports and set imports in self
        if imports:
            self.imports = []
            syspaths = syspaths if syspaths else []
            for i in imports:
                if i.startswith('sys.path.append')\
                        or i.startswith('sys.path.insert'):
                    syspaths.append(i)
                else:
                    self.imports.append(i)

        # Function specific initialization
        if callable(command):
            self.kind = 'function'
            script_file = _os.path.join(
                self.scriptpath, '{}_func.{}.py'.format(name, suffix)
                )
            self.poutfile = self.outfile + '.func.pickle'
            self.function = _Function(
                file_name=script_file, function=command, args=args,
                kwargs=kwargs, imports=self.imports, syspaths=syspaths,
                outfile=self.poutfile
            )
            # Collapse the _command into a python call to the function script
            executable = '#!/usr/bin/env python{}'.format(
                _sys.version_info.major) if _conf.get_option(
                    'jobs', 'generic_python') else _sys.executable

            command = '{} {}'.format(executable, self.function.file_name)
            args = None
        else:
            self.kind = 'script'
            self.poutfile = None

        # Collapse args into command
        command = command + ' '.join(args) if args else command

        #####################
        #  Script Creation  #
        #####################

        # Build execution wrapper with modules
        precmd  = ''
        if self.modules:
            for module in self.modules:
                precmd += 'module load {}\n'.format(module)

        # Create queue-dependent scripts
        sub_script = ''
        if self.qtype == 'slurm':
            scrpt = _os.path.join(
                self.scriptpath, '{}.{}.sbatch'.format(name, suffix)
            )

            # We use a separate script and a single srun command to avoid
            # issues with multiple threads running at once
            exec_script  = _os.path.join(self.scriptpath,
                                         '{}.{}.script'.format(name, suffix))
            exe_script   = _scrpts.CMND_RUNNER_TRACK.format(
                precmd=precmd, usedir=self.runpath, name=name, command=command)
            # Create the exec_script Script object
            self.exec_script = _Script(script=exe_script,
                                       file_name=exec_script)

            # Add all of the keyword arguments at once
            precmd = _options.options_to_string(kwds, self.qtype) + precmd

            ecmnd = 'srun bash {}'.format(exec_script)
            sub_script = _scrpts.SCRP_RUNNER.format(precmd=precmd,
                                                    script=exec_script,
                                                    command=ecmnd)

        elif self.qtype == 'torque':
            scrpt = _os.path.join(self.scriptpath,
                                  '{}.cluster.qsub'.format(name))

            # Add all of the keyword arguments at once
            precmd = _options.options_to_string(kwds, self.qtype) + precmd

            sub_script = _scrpts.CMND_RUNNER_TRACK.format(
                precmd=precmd, usedir=self.runpath, name=name, command=command)

        elif self.qtype == 'local':
            # Create the pool
            if not _local.JQUEUE or not _local.JQUEUE.runner.is_alive():
                threads = kwds['threads'] if 'threads' in kwds \
                        else _local.THREADS
                _local.JQUEUE = _local.JobQueue(cores=threads)

            scrpt = _os.path.join(self.scriptpath, '{}.cluster'.format(name))
            sub_script = _scrpts.CMND_RUNNER_TRACK.format(
                precmd=precmd, usedir=self.runpath, name=name, command=command)

        else:
            raise _ClusterError('Invalid queue type')

        # Create the submission Script object
        self.submission = _Script(script=sub_script,
                                  file_name=scrpt)

        # Save the keyword arguments for posterity
        self.kwargs = kwds

    ####################
    #  Public Methods  #
    ####################

    def write(self, overwrite=True):
        """Write all scripts.

        Args:
            overwrite (bool): Overwrite existing files, defaults to True.
        """
        _logme.log('Writing files, overwrite={}'.format(overwrite), 'debug')
        self.submission.write(overwrite)
        if self.exec_script:
            self.exec_script.write(overwrite)
        if self.function:
            self.function.write(overwrite)
        self.written = True

    def clean(self, delete_outputs=None, get_outputs=True):
        """Delete all scripts created by this module, if they were written.

        Args:
            delete_outputs (bool): also delete all output and err files,
                                   but get their contents first.
            get_outputs (bool):    if delete_outputs, save outputs before
                                   deleting.
        """
        _logme.log('Cleaning outputs, delete_outputs={}'
                   .format(delete_outputs), 'debug')
        if delete_outputs is None:
            delete_outputs = self.clean_outputs
        assert isinstance(delete_outputs, bool)
        for jobfile in [self.submission, self.exec_script, self.function]:
            if jobfile:
                jobfile.clean()
        if delete_outputs:
            _logme.log('Deleting output files.', 'debug')
            if get_outputs:
                self.fetch_outputs(delete_files=True)
            for f in self.outfiles:
                if _os.path.isfile(f):
                    _logme.log('Deleteing {}'.format(f), 'debug')
                    _os.remove(f)

    def submit(self, wait_on_max_queue=True):
        """Submit this job.

        Args:
            wait_on_max_queue (bool): Block until queue limit is below the
                                      maximum before submitting.

        To disable max_queue_len, set it to 0. None will allow override by
        the default settings in the config file, and any positive integer will
        be interpretted to be the maximum queue length.

        Returns:
            self
        """
        if self.submitted:
            _logme.log('Not submitting, already submitted.', 'warn')
            return self

        if not self.written:
            self.write()

        # Check dependencies
        dependencies = []
        if self.dependencies:
            for depend in self.dependencies:
                if isinstance(depend, Job):
                    if not depend.id:
                        _logme.log(
                            'Cannot submit job as dependency {} '
                            .format(depend) + 'has not been submitted',
                            'error'
                        )
                        return self
                    dependencies.append(int(depend.id))
                else:
                    dependencies.append(int(depend))

        # Wait on the queue if necessary
        if wait_on_max_queue:
            self.update()
            self.queue.wait_to_submit()

        # Only include queued or running dependencies
        self.queue._update()  # Force update
        depends = []
        for depend in dependencies:
            dep_check = self.queue.check_dependencies(depend)
            if dep_check == 'absent':
                _logme.log(
                    'Cannot submit job as dependency {} '
                    .format(depend) + 'is not in the queue',
                    'error'
                )
                return self
            elif dep_check == 'good':
                _logme.log(
                    'Dependency {} is complete, skipping'
                    .format(depend), 'debug'
                )
            elif dep_check == 'bad':
                _logme.log(
                    'Cannot submit job as dependency {} '
                    .format(depend) + 'has failed',
                    'error'
                )
                return self
            elif dep_check == 'active':
                if self.queue.jobs[depend].state == 'completeing':
                    continue
                _logme.log('Dependency {} is {}, adding to deps'
                           .format(depend, self.queue.jobs[depend].state),
                           'debug')
                depends.append(depend)
            else:
                # This shouldn't happen ever
                raise _ClusterError('fyrd.queue.Queue.check_dependencies() ' +
                                    'returned an unrecognized value {}'
                                    .format(dep_check))

        if self.qtype == 'local':
            # Normal mode dependency tracking uses only integer job numbers
            _logme.log('Submitting to local', 'debug')
            command = 'bash {}'.format(self.submission.file_name)
            fileargs  = dict(stdout=self.outfile,
                             stderr=self.errfile)

            # Make sure the global job pool exists
            if not _local.JQUEUE or not _local.JQUEUE.runner.is_alive():
                _local.JQUEUE = _local.JobQueue(cores=_local.THREADS)
            self.id = _local.JQUEUE.add(_run.cmd, args=(command,),
                                        kwargs=fileargs,
                                        dependencies=depends,
                                        cores=self.cores)
            self.submitted = True
            self.submit_time = _dt.now()
            self.state = 'submitted'

        elif self.qtype == 'slurm':
            _logme.log('Submitting to slurm', 'debug')
            if self.depends:
                deps = '--dependency=afterok:{}'.format(
                    ':'.join([str(d) for d in depends]))
                args = ['sbatch', deps, self.submission.file_name]
            else:
                args = ['sbatch', self.submission.file_name]

            # Try to submit job 5 times
            code, stdout, stderr = _run.cmd(args, tries=5)
            if code == 0:
                self.id = int(stdout.split(' ')[-1])
            else:
                _logme.log('sbatch failed with code {}\n'.format(code) +
                           'stdout: {}\nstderr: {}'.format(stdout, stderr),
                           'critical')
                raise _CalledProcessError(code, args, stdout, stderr)
            self.submitted = True
            self.submit_time = _dt.now()
            self.state = 'submitted'

        elif self.qtype == 'torque':
            _logme.log('Submitting to torque', 'debug')
            if self.depends:
                deps = '-W depend={}'.format(
                    ','.join(['afterok:' + str(d) for d in depends]))
                args = ['qsub', deps, self.submission.file_name]
            else:
                args = ['qsub', self.submission.file_name]

            # Try to submit job 5 times
            code, stdout, stderr = _run.cmd(args, tries=5)
            if code == 0:
                self.id = int(stdout.split('.')[0])
            elif code == 17 and 'Unable to open script file' in stderr:
                _logme.log('qsub submission failed due to an already existing '
                           'script file, attempting to rename file and try '
                           'again.\nstderr: {}, stdout: {}, cmnd: {}'
                           .format(stderr, stdout, args), 'error')
                new_name = args[1] + '.resub'
                _os.rename(args[1], new_name)
                _logme.log('renamed script {} to {}, resubmitting'
                           .format(args[1], new_name), 'info')
                args[1] = new_name
                code, stdout, stderr = _run.cmd(args, tries=5)
                if code == 0:
                    self.id = int(stdout.split('.')[0])
                else:
                    _logme.log('Resubmission still failed, aborting',
                               'critical')
                    raise _CalledProcessError(code, args, stdout, stderr)
            else:
                if stderr.startswith('qsub: submit error ('):
                    raise _ClusterError('qsub submission failed with error: ' +
                                        '{}, command: {}'.format(stderr, args))
                else:
                    _logme.log(
                        'qsub failed with code {}\nstdout: {}\nstderr: {}'
                        .format(code, stdout, stderr), 'critical'
                    )
                    raise _CalledProcessError(code, args, stdout, stderr)
            self.submitted = True
            self.submit_time = _dt.now()
            self.state = 'submitted'
        else:
            raise _ClusterError("Invalid queue type {}".format(self.qtype))

        if not self.submitted:
            raise _ClusterError('Submission appears to have failed, this '
                                "shouldn't happen")

        return self

    def resubmit(self):
        """Attempt to auto resubmit, deletes prior files."""
        self.clean(delete_outputs=True)
        self.state = 'Not_Submitted'
        self.write()
        return self.submit()

    def wait(self):
        """Block until job completes."""
        if not self.submitted:
            if _conf.get_option('jobs', 'auto_submit'):
                _logme.log('Auto-submitting as not submitted yet', 'debug')
                self.submit()
                _sleep(0.5)
            else:
                _logme.log('Cannot wait for result as job has not been ' +
                           'submitted', 'warn')
                return False
        _sleep(0.1)
        self.update(False)
        if not self.done:
            _logme.log('Waiting for self {}'.format(self.name), 'debug')
            status = self.queue.wait(self.id)
            if status == 'disappeared':
                self.state = status
            elif status is not True:
                return False
            else:
                self.update()
                if self.get_exitcode(update=False) != 0:
                    _logme.log('Job failed with exitcode {}'
                               .format(self.exitcode), 'debug')
                    return False
        if self.wait_for_files(caution_message=False):
            self.update()
            if self.state == 'disappeared':
                _logme.log('Job files found for disappered job, assuming '
                           'success', 'info')
                return 'disappeared'
            return True
        else:
            if self.state == 'disappeared':
                _logme.log('Disappeared job has no output files, assuming '
                           'failure', 'error')
            return False

    def wait_for_files(self, btme=None, caution_message=False):
        """Block until files appear up to 'file_block_time' in config file.

        Aborts after 2 seconds if job exit code is not 0.

        Args:
            btme (int):             Number of seconds to try for before giving
                                    up, default set in config file.
            caution_message (bool): Display a message if this is taking
                                    a while.

        Returns:
            bool: True if files found
        """
        if not self.done:
            _logme.log("Cannot wait for files if we aren't complete",
                       'warn')
            return False
        wait_time = 0.1 # seconds
        if btme:
            lvl = 'debug'
        else:
            lvl = 'warn'
            btme = _conf.get_option('jobs', 'file_block_time')
        start = _dt.now()
        dsp   = False
        _logme.log('Checking for output files', 'debug')
        while True:
            runtime = (_dt.now() - start).seconds
            if caution_message and runtime > 1:
                _logme.log('Job complete.', 'info')
                _logme.log('Waiting for output files to appear.', 'info')
                caution_message = False
            if not dsp and runtime > 20:
                _logme.log('Still waiting for output files to appear',
                           'info')
                dsp = True
            count = 0
            outfiles = self.incomplete_outfiles
            tlen = len(outfiles)
            if not outfiles:
                _logme.log('No incomplete outfiles, assuming all found in ' +
                           '{} seconds'.format(runtime), 'debug')
                break
            for i in outfiles:
                if _os.path.isfile(i):
                    count += 1
            if count == tlen:
                _logme.log('All output files found in {} seconds'
                           .format(runtime), 'debug')
                break
            _sleep(wait_time)
            if runtime > btme:
                _logme.log('Job completed but files have not appeared for ' +
                           '>{} seconds'.format(btme), lvl)
                return False
            self.update()
            if runtime > 2 and self.get_exitcode(update=False) != 0:
                _logme.log('Job failed with exit code {}.'
                           .format(self.exitcode) + ' Cannot find files.',
                           'error')
                return False
            if _queue.MODE == 'local':
                _logme.log('Job output files were not found.', 'error')
                _logme.log('Expected files: {}'.format(self.outfiles))
                return False
        return True

    def get(self, save=True, cleanup=None, delete_outfiles=None,
            del_no_save=None, raise_on_error=True):
        """Block until job completed and return output of script/function.

        By default saves all outputs to this class and deletes all intermediate
        files.

        Args:
            save (bool):            Save all outputs to the class also (advised)
            cleanup (bool):         Clean all intermediate files after job
                                    completes.
            delete_outfiles (bool): Clean output files after job completes.
            del_no_save (bool):     Delete output files even if `save` is
                                    `False`
            raise_on_error (bool): If the returned output is an Exception,
                                   raise it.

        Returns:
            str: Function output if Function, else STDOUT
        """
        _logme.log(('Getting outputs, cleanup={}, autoclean={}, '
                    'delete_outfiles={}').format(
                        cleanup, self.clean_files, delete_outfiles
                    ), 'debug')
        # Wait for queue
        status = self.wait()
        if status == 'disappeared':
            _logme.log('Job disappeared from queue, attempting to get outputs',
                       'debug')
            try:
                self.fetch_outputs(save=save, delete_files=False,
                                   get_stats=False)
            except IOError:
                _logme.log('Job disappeared from the queue and files could not'
                           ' be found, job must have died and been deleted '
                           'from the queue', 'critical')
                raise IOError('Job {} disappeared, output files missing'
                              .format(self))
        elif status is not True:
            _logme.log('Wait failed, cannot get outputs, aborting', 'error')
            self.update()
            if _os.path.isfile(self.errfile):
                if _logme.MIN_LEVEL in ['debug', 'verbose']:
                    _sys.stderr.write('STDERR of Job:\n')
                    _sys.stderr.write(self.get_stderr(delete_file=False,
                                                      update=False))
            if self.poutfile and _os.path.isfile(self.poutfile):
                _logme.log('Getting pickled output', 'debug')
                self.get_output(delete_file=False, update=False,
                                raise_on_error=raise_on_error)
            else:
                _logme.log('Pickled out file does not exist, cannot get error',
                           'debug')
            return
        else:
            # Get output
            _logme.log('Wait complete, fetching outputs', 'debug')
            self.fetch_outputs(save=save, delete_files=False)
        out = self.out if save else self.get_output(save=save)
        # Cleanup
        if cleanup is None:
            cleanup = self.clean_files
        else:
            assert isinstance(cleanup, bool)
        if delete_outfiles is None:
            delete_outfiles = self.clean_outputs
        if save is False:
            delete_outfiles = del_no_save if del_no_save is not None else False
        if cleanup:
            self.clean(delete_outputs=delete_outfiles)
        return out

    def get_output(self, save=True, delete_file=None, update=True,
                   raise_on_error=True):
        """Get output of function or script.

        This is the same as stdout for a script, or the function output for
        a function.

        By default, output file is kept unless delete_file is True or
        self.clean_files is True.

        Args:
            save (bool):           Save the output to self.out, default True.
                                   Would be a good idea to set to False if the
                                   output is huge.
            delete_file (bool):    Delete the output file when getting
            update (bool):         Update job info from queue first.
            raise_on_error (bool): If the returned output is an Exception,
                                   raise it.

        Returns:
            The output of the script or function. Always a string if script.
        """
        _logme.log(('Getting output, save={}, clean_files={}, '
                    'delete_file={}').format(
                        save, self.clean_files, delete_file
                    ), 'debug')
        if delete_file is None:
            delete_file = self.clean_outputs
        if self.kind == 'script':
            return self.get_stdout(save=save, delete_file=delete_file,
                                   update=update)
        if self.done and self._got_out:
            _logme.log('Getting output from _out', 'debug')
            return self._out
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self.wait_for_files()
        else:
            _logme.log('Cannot get pickled output before job completes',
                       'warn')
            return None
        _logme.log('Getting output from {}'.format(self.poutfile), 'debug')
        if _os.path.isfile(self.poutfile):
            with open(self.poutfile, 'rb') as fin:
                out = _pickle.load(fin)
            if delete_file is True or self.clean_files is True:
                _logme.log('Deleting {}'.format(self.poutfile),
                           'debug')
                _os.remove(self.poutfile)
            if save:
                self._out = out
                self._got_out = True
            if _run.is_exc(out):
                _logme.log('{} failed with exception {}'.format(self, out[1]),
                           'error')
                if raise_on_error:
                    _reraise(*out)
            return out
        else:
            _logme.log('No file at {} even though job has completed!'
                       .format(self.poutfile), 'critical')
            raise IOError('File not found: {}'.format(self.poutfile))

    def get_stdout(self, save=True, delete_file=None, update=True):
        """Get stdout of function or script, same for both.

        By default, output file is kept unless delete_file is True or
        self.clean_files is True.

        Args:
            save (bool):        Save the output to self.stdout, default True.
                                Would be a good idea to set to False if the
                                output is huge.
            delete_file (bool): Delete the stdout file when getting
            update (bool):      Update job info from queue first.

        Returns:
            str: The contents of STDOUT, with runtime info and trailing
                 newline removed.

        Also sets self.start and self.end from the contents of STDOUT if
        possible.
        """
        if delete_file is None:
            delete_file = self.clean_outputs
        _logme.log(('Getting stdout, save={}, clean_files={}, '
                    'delete_file={}').format(
                        save, self.clean_files, delete_file
                    ), 'debug')
        if self.done and self._got_stdout:
            _logme.log('Getting stdout from _stdout', 'debug')
            return self._stdout
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self.wait_for_files()
        else:
            _logme.log('Job not done, attempting to get current STDOUT ' +
                       'anyway', 'info')
        _logme.log('Getting stdout from {}'.format(self.kwargs['outfile']),
                   'debug')
        if _os.path.isfile(self.kwargs['outfile']):
            self.get_times(update=False)
            stdout = open(self.kwargs['outfile']).read()
            if stdout:
                stdouts = stdout.split('\n')
                stdout     = '\n'.join(stdouts[2:-3]) + '\n'
            if delete_file is True or self.clean_files is True:
                _logme.log('Deleting {}'.format(self.kwargs['outfile']),
                           'debug')
                _os.remove(self.kwargs['outfile'])
            if save:
                self._stdout = stdout
                if self.done:
                    self._got_stdout = True
            return stdout
        else:
            _logme.log('No file at {}, cannot get stdout'
                       .format(self.kwargs['outfile']), 'warn')
            return None

    def get_stderr(self, save=True, delete_file=None, update=True):
        """Get stderr of function or script, same for both.

        By default, output file is kept unless delete_file is True or
        self.clean_files is True.

        Args:
            save (bool):        Save the output to self.stdout, default True.
                                Would be a good idea to set to False if the
                                output is huge.
            delete_file (bool): Delete the stdout file when getting
            update (bool):      Update job info from queue first.

        Returns:
            str: The contents of STDERR, with trailing newline removed.
        """
        if delete_file is None:
            delete_file = self.clean_outputs
        _logme.log(('Getting stderr, save={}, clean_files={}, '
                    'delete_file={}').format(
                        save, self.clean_files, delete_file
                    ), 'debug')
        if self.done and self._got_stderr:
            _logme.log('Getting stderr from _stderr', 'debug')
            return self._stderr
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self.wait_for_files()
        else:
            _logme.log('Job not done, attempting to get current STDERR ' +
                       'anyway', 'info')
        _logme.log('Getting stderr from {}'.format(self.kwargs['errfile']),
                   'debug')
        if _os.path.isfile(self.kwargs['errfile']):
            stderr = open(self.kwargs['errfile']).read()
            if delete_file is True or self.clean_files is True:
                _logme.log('Deleting {}'.format(self.kwargs['errfile']),
                           'debug')
                _os.remove(self.kwargs['errfile'])
            if save:
                self._stderr = stderr
                if self.done:
                    self._got_stderr = True
            return stderr
        else:
            _logme.log('No file at {}, cannot get stderr'
                       .format(self.kwargs['errfile']), 'warn')
            return None

    def get_times(self, update=True):
        """Get stdout of function or script, same for both.

        Args:
            update (bool): Update job info from queue first.

        Returns:
            tuple: start, end as two datetime objects.

        Also sets self.start and self.end from the contents of STDOUT if
        possible.
        """
        _logme.log('Getting times', 'debug')
        if self.done and self._got_times:
            _logme.log('Getting times from self.start, self.end', 'debug')
            return self.start, self.end
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self.wait_for_files()
        else:
            _logme.log('Cannot get times until job is complete.', 'warn')
            return None, None
        _logme.log('Getting times from {}'.format(self.kwargs['outfile']),
                   'debug')
        if _os.path.isfile(self.kwargs['outfile']):
            stdout = open(self.kwargs['outfile']).read()
            if stdout:
                stdouts = stdout.split('\n')
                # Get times
                timefmt = '%y-%m-%d-%H:%M:%S'
                try:
                    self.start = _dt.strptime(stdouts[0], timefmt)
                    self.end   = _dt.strptime(stdouts[-2], timefmt)
                except ValueError as err:
                    _logme.log('Time parsing failed with value error; ' +
                               '{}. '.format(err) + 'This may be because you ' +
                               'are using the script running that does not ' +
                               'include time tracking', 'debug')
            self._got_times = True
            return self.start, self.end
        else:
            _logme.log('No file at {}, cannot get times'
                       .format(self.kwargs['outfile']), 'warn')
            return None

    def get_exitcode(self, update=True):
        """Try to get the exitcode.

        Args:
            update (bool): Update job info from queue first.

        Returns:
            int: The exitcode of the running process.
        """
        _logme.log('Getting exitcode', 'debug')
        if self.done and self._got_exitcode:
            _logme.log('Getting exitcode from _exitcode', 'debug')
            return self._exitcode
        if self.status == 'disappeared':
            _logme.log('Cannot get exitcode for disappeared job', 'debug')
            return 0
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self.wait_for_files()
        else:
            _logme.log('Job is not complete, no exit code yet', 'info')
            return None
        _logme.log('Getting exitcode from queue', 'debug')
        if not self.queue_info:
            self.queue_info = self.queue[self.id]
        if hasattr(self.queue_info, 'exitcode'):
            code = self.queue_info.exitcode
        else:
            code = None
            _logme.log('No exitcode even though the job is done, this ' +
                       "shouldn't happen.", 'warn')
        self._exitcode = code
        self._got_exitcode = True
        if code != 0:
            self.state = 'failed'
            _logme.log('Job {} failed with exitcode {}'
                       .format(self.name, code), 'error')
        return code

    def update(self, fetch_info=True):
        """Update status from the queue.

        Args:
            fetch_info (bool): Fetch basic job info if complete.
        """
        if not self._updating:
            self._update(fetch_info)
        else:
            _logme.log('Already updating, aborting.', 'debug')

    def update_queue_info(self):
        """Set queue_info from the queue even if done."""
        _logme.log('Updating queue_info', 'debug')
        queue_info1 = self.queue[self.id]
        self.queue.update()
        queue_info2 = self.queue[self.id]
        if queue_info2:
            self.queue_info = queue_info2
        elif queue_info1:
            self.queue_info = queue_info1
        elif self.queue_info is None and self.submitted:
            _logme.log('Cannot find self in the queue and queue_info is empty',
                       'warn')
        return self.queue_info

    def fetch_outputs(self, save=True, delete_files=None, get_stats=True):
        """Save all outputs in their current state. No return value.

        This method does not wait for job completion, but merely gets the
        outputs. To wait for job completion, use `get()` instead.

        Args:
            save (bool):         Save all outputs to the class also (advised)
            delete_files (bool): Delete the output files when getting, only
                                 used if save is True
            get_stats (bool):    Try to get exitcode.
        """
        _logme.log('Saving outputs to self, delete_files={}'
                   .format(delete_files), 'debug')
        self.update()
        if delete_files is None:
            delete_files = self.clean_outputs
        if not self._got_exitcode and get_stats:
            self.get_exitcode(update=False)
        if not self._got_times:
            self.get_times(update=False)
        if save:
            self.get_output(save=True, delete_file=delete_files, update=False)
            self.get_stdout(save=True, delete_file=delete_files, update=False)
            self.get_stderr(save=True, delete_file=delete_files, update=False)

    @property
    def files(self):
        """Build a list of files associated with this class."""
        files = [self.submission]
        if self.kind == 'script':
            files.append(self.exec_script)
        if self.kind == 'function':
            files.append(self.function)
        return files

    @property
    def runtime(self):
        """Return the runtime."""
        if not self.done:
            _logme.log('Cannot get runtime as not yet complete.' 'warn')
            return None
        if not self.start:
            self.get_times()
        return self.end-self.start

    @property
    def done(self):
        """Check if completed or not.

        Updates the Job and Queue.

        Returns:
            Bool: True if complete, False otherwise.
        """
        # We have the same statement twice to try and avoid updating.
        if self.state in _queue.DONE_STATES:
            return True
        if not self._updating:
            self.update()
            if self.state in _queue.DONE_STATES:
                return True
        return False

    ###############
    #  Internals  #
    ###############

    def _update(self, fetch_info=True):
        """Update status from the queue.

        Args:
            fetch_info (bool): Fetch basic job info if complete.
        """
        _logme.log('Updating job.', 'debug')
        self._updating = True
        if self.done or not self.submitted:
            self._updating = False
            return
        self.queue.update()
        if self.submitted and self.id:
            queue_info = self.queue[self.id]
            if queue_info:
                assert self.id == queue_info.id
                self.found = True
                self.queue_info = queue_info
                self.state = self.queue_info.state
                if self.done and fetch_info:
                    if self.wait_for_files(btme=1, caution_message=False):
                        if not self._got_exitcode:
                            self.get_exitcode(update=False)
                        if not self._got_times:
                            self.get_times(update=False)
            elif self.found:
                _logme.log('Job appears to have disappeared, waiting for '
                           'reappearance, this may take a while', 'warn')
                status = self.wait()
                if status == 'disappeared':
                    _logme.log('Job disappeared, but the output files are '
                               'present', 'info')
                elif not status:
                    _logme.log('Job appears to have failed and disappeared',
                               'error')
            # If job not found after 360 seconds, assume trouble
            elif self.submitted and (_dt.now()-self.submit_time).seconds > 1000:
                s = (_dt.now()-self.submit_time).seconds
                _logme.log('Job not in queue after {} seconds of searching.'
                           .format(s), 'warn')
        self._updating = False

    @property
    def outfiles(self):
        """A list of all outfiles associated with this Job."""
        outfiles = [self.outfile, self.errfile]
        if self.poutfile:
            outfiles.append(self.poutfile)
        return outfiles

    @property
    def incomplete_outfiles(self):
        """A list of all outfiles that haven't already been fetched."""
        outfiles = []
        if self.outfile and not self._got_stdout:
            outfiles.append(self.outfile)
        if self.errfile and not self._got_stderr:
            outfiles.append(self.errfile)
        if self.poutfile and not self._got_out:
            outfiles.append(self.poutfile)
        return outfiles

    def __getattr__(self, key):
        """Dynamically get out, stdout, stderr, and exitcode."""
        if key == 'out':
            return self.get_output()
        elif key == 'stdout':
            return self.get_stdout()
        elif key == 'stderr':
            return self.get_stderr()
        elif key == 'exitcode':
            return self.get_exitcode()
        elif key == 'err':
            return self.get_stderr()

    def __repr__(self):
        """Return simple job information."""
        outstr = "Job:{name}<{mode}:{qtype}".format(
            name=self.name, mode=self.kind, qtype=self.qtype)
        if self.submitted:
            outstr += ':{}'.format(self.id)
        outstr += "(command:{cmnd})".format(cmnd=self.command)
        if self.submitted or self.done:
            outstr += self.state.upper()
        elif self.written:
            outstr += "WRITTEN"
        else:
            outstr += "NOT_SUBMITTED"
        outstr += ">"
        return outstr

    def __str__(self):
        """Print job name and ID + status."""
        if self.done:
            state = 'completed'
            id1 = str(self.id)
        elif self.written:
            state = 'written'
            id1 = str(self.id)
        else:
            state = 'not written'
            id1 = 'NA'
        return "Job: {name} ID: {id}, state: {state}".format(
            name=self.name, id=id1, state=state)



# -*- coding: utf-8 -*-
"""
Functions to allow simple job and file submission without the Job class.
"""
import os  as _os
import sys as _sys
from time import sleep as _sleep
from subprocess import CalledProcessError as _CalledProcessError

###############################################################################
#                               Import Ourself                                #
###############################################################################


from . import run   as _run
from . import conf  as _conf
from . import queue as _queue
from . import local as _local
from . import logme as _logme
from . import ClusterError as _ClusterError
from .job import Job

__all__ = ['submit', 'make_job', 'make_job_file', 'submit_file', 'clean_dir']

###############################################################################
#                            Submission Functions                             #
###############################################################################


def submit(command, args=None, kwargs=None, name=None, qtype=None,
           profile=None, **kwds):
    """Submit a script to the cluster.

    Args:
            command (function/str): The command or function to execute.
            args (tuple/dict):      Optional arguments to add to command,
                                    particularly useful for functions.
            kwargs (dict):          Optional keyword arguments to pass to the
                                    command, only used for functions.
            name (str):             Optional name of the job. If not defined,
                                    guessed. If a job of the same name is
                                    already queued, an integer job number (not
                                    the queue number) will be added, ie.
                                    <name>.1
            qtype (str):            Override the default queue type
            profile (str):          The name of a profile saved in the
                                    conf

            *All other keywords are parsed into cluster keywords by the
            options system. For available keywords see `fyrd.option_help()`*

    Returns:
        Job object
    """

    _queue.check_queue()  # Make sure the queue.MODE is usable

    job = Job(command=command, args=args, kwargs=kwargs, name=name,
              qtype=qtype, profile=profile, **kwds)

    job.write()
    job.submit()
    job.update()

    return job


#########################
#  Job file generation  #
#########################


def make_job(command, args=None, kwargs=None, name=None, qtype=None,
             profile=None, **kwds):
    """Make a job file compatible with the chosen cluster.

    If mode is local, this is just a simple shell script.

    Args:
            command (function/str): The command or function to execute.
            args (tuple/dict):      Optional arguments to add to command,
                                    particularly useful for functions.
            kwargs (dict):          Optional keyword arguments to pass to the
                                    command, only used for functions.
            name (str):             Optional name of the job. If not defined,
                                    guessed. If a job of the same name is
                                    already queued, an integer job number (not
                                    the queue number) will be added, ie.
                                    <name>.1
            qtype (str):            Override the default queue type
            profile (str):          The name of a profile saved in the
                                    conf

            *All other keywords are parsed into cluster keywords by the
            options system. For available keywords see `fyrd.option_help()`*

    Returns:
        Job object
    """

    _queue.check_queue()  # Make sure the queue.MODE is usable

    job = Job(command=command, args=args, kwargs=kwargs, name=name,
              qtype=qtype, profile=profile, **kwds)

    # Return the path to the script
    return job


def make_job_file(command, args=None, kwargs=None, name=None, qtype=None,
                  profile=None, **kwds):
    """Make a job file compatible with the chosen cluster.

    Args:
            command (function/str): The command or function to execute.
            args (tuple/dict):      Optional arguments to add to command,
                                    particularly useful for functions.
            kwargs (dict):          Optional keyword arguments to pass to the
                                    command, only used for functions.
            name (str):             Optional name of the job. If not defined,
                                    guessed. If a job of the same name is
                                    already queued, an integer job number (not
                                    the queue number) will be added, ie.
                                    <name>.1
            qtype (str):            Override the default queue type
            profile (str):          The name of a profile saved in the
                                    conf

            *All other keywords are parsed into cluster keywords by the
            options system. For available keywords see `fyrd.option_help()`*

    Returns:
        Job object
    """

    _queue.check_queue()  # Make sure the queue.MODE is usable

    job = Job(command=command, args=args, kwargs=kwargs, name=name,
              qtype=qtype, profile=profile, **kwds)

    job = job.write()

    # Return the path to the script
    return job.submission


##############
#  Cleaning  #
##############


def clean(jobs):
    """Delete all files in jobs list or single Job object."""
    if isinstance(jobs, Job):
        jobs = [jobs]
    if not isinstance(jobs, (list, tuple)):
        raise _ClusterError('Job list must be a Job, list, or tuple')
    for job in jobs:
        job.clean()


###############################################################################
#                      Job Object Independent Functions                       #
###############################################################################


def submit_file(script_file, dependencies=None, threads=None, qtype=None):
    """Submit a job file to the cluster.

    If qtype or queue.MODE is torque, qsub is used; if it is slurm, sbatch
    is used; if it is local, the file is executed with subprocess.

    This function is independent of the Job object and just submits a file.

    Args:
        dependencies: A job number or list of job numbers.
                      In slurm: `--dependency=afterok:` is used
                      For torque: `-W depend=afterok:` is used

        threads:      Total number of threads to use at a time, defaults to all.
                      ONLY USED IN LOCAL MODE

    Returns:
        job number for torque or slurm multiprocessing job object for local
        mode
    """
    _queue.check_queue()  # Make sure the queue.MODE is usable

    if not qtype:
        qtype = _queue.get_cluster_environment()

    # Check dependencies
    if dependencies:
        if isinstance(dependencies, (str, int)):
            dependencies = [dependencies]
        if not isinstance(dependencies, (list, tuple)):
            raise Exception('dependencies must be a list, int, or string.')
        dependencies = [str(i) for i in dependencies]

    if qtype == 'slurm':
        if dependencies:
            dependencies = '--dependency=afterok:{}'.format(
                ':'.join([str(d) for d in dependencies]))
            args = ['sbatch', dependencies, script_file]
        else:
            args = ['sbatch', script_file]
        # Try to submit job 5 times
        count = 0
        while True:
            code, stdout, stderr = _run.cmd(args)
            if code == 0:
                job = int(stdout.split(' ')[-1])
                break
            else:
                if count == 5:
                    _logme.log('sbatch failed with code {}\n'.format(code),
                               'stdout: {}\nstderr: {}'.format(stdout, stderr),
                               'critical')
                    raise _CalledProcessError(code, args, stdout, stderr)
                _logme.log('sbatch failed with err {}. Resubmitting.'.format(
                    stderr), 'debug')
                count += 1
                _sleep(1)
                continue
            break
        return job

    elif qtype == 'torque':
        if dependencies:
            dependencies = '-W depend={}'.format(
                ','.join(['afterok:' + d for d in dependencies]))
            args = ['qsub', dependencies, script_file]
        else:
            args = ['qsub', script_file]
        # Try to submit job 5 times
        count = 0
        while True:
            code, stdout, stderr = _run.cmd(args)
            if code == 0:
                job = int(stdout.split('.')[0])
                break
            else:
                if count == 5:
                    _logme.log('qsub failed with code {}\n'.format(code),
                               'stdout: {}\nstderr: {}'.format(stdout, stderr),
                               'critical')
                    raise _CalledProcessError(code, args, stdout, stderr)
                _logme.log('qsub failed with err {}. Resubmitting.'.format(
                    stderr), 'debug')
                count += 1
                _sleep(1)
                continue
            break
        return job

    elif qtype == 'local':
        # Normal mode dependency tracking uses only integer job numbers
        depends = []
        if dependencies:
            for depend in dependencies:
                if isinstance(depend, Job):
                    depends.append(int(depend.id))
                else:
                    depends.append(int(depend))
        command = 'bash {}'.format(script_file)
        # Make sure the global job pool exists
        if not _local.JQUEUE or not _local.JQUEUE.runner.is_alive():
            _local.JQUEUE = _local.JobQueue(cores=threads)
        return _local.JQUEUE.add(_run.cmd, (command,), dependencies=depends)


def clean_work_dirs(outputs=False, confirm=False):
    """Clean all files in the scriptpath and outpath directories.

    Args:
        outputs (bool): Also delete output files.
        confirm (bool): Confirm on command line before deleteing.

    Returns:
        list: A list of deleted files.
    """
    files = []
    scriptpath = _conf.get_option('jobs', 'scriptpath')
    outpath    = _conf.get_option('jobs', 'outpath')
    if scriptpath:
        files += clean_dir(scriptpath, delete_outputs=outputs, confirm=confirm)
    if outputs and outpath and outpath != scriptpath:
        files += clean_dir(outpath, delete_outputs=True, confirm=confirm)
    return files


def clean_dir(directory=None, suffix=None, qtype=None, confirm=False,
              delete_outputs=None):
    """Delete all files made by this module in directory.

    CAUTION: The clean() function will delete **EVERY** file with
             extensions matching those these::
                 .<suffix>.err
                 .<suffix>.out
                 .<suffix>.out.func.pickle
                 .<suffix>.sbatch & .<suffix>.script for slurm mode
                 .<suffix>.qsub for torque mode
                 .<suffix> for local mode
                 _func.<suffix>.py
                 _func.<suffix>.py.pickle.in
                 _func.<suffix>.py.pickle.out

    Args:
        directory (str):       The directory to run in, defaults to the current
                               directory.
        suffix (str):          Override the default suffix.
        qtype (str):           Only run on files of this qtype
        confirm (bool):        Ask the user before deleting the files
        delete_outputs (bool): Delete all output files too.

    Returns:
        A set of deleted files
    """
    _queue.check_queue(qtype)  # Make sure the queue.MODE is usable

    if delete_outputs is None:
        delete_outputs = _conf.get_option('jobs', 'clean_outputs')

    # Sanitize arguments
    directory = _os.path.abspath(directory if directory else '.')
    suffix = suffix if suffix else _conf.get_option('jobs', 'suffix')

    # Extension patterns to delete
    extensions = ['_func.' + suffix + '.py']
    if delete_outputs:
        extensions += ['.' + suffix + '.err', '.' + suffix + '.out',
                       '_func.' + suffix + '.py.pickle.out',
                       '.' + suffix + '.out.func.pickle']

    if qtype:
        if qtype == 'local':
            extensions.append('.' + suffix)
        elif qtype == 'slurm':
            extensions += ['.' + suffix + '.sbatch', '.' + suffix + '.script']
        elif qtype== 'torque':
            extensions.append('.' + suffix + '.qsub')
    else:
        extensions.append('.' + suffix)
        extensions.append('_func.' + suffix + '.py.pickle.in')
        extensions += ['.' + suffix + '.sbatch', '.' + suffix + '.script']
        extensions.append('.' + suffix + '.qsub')

    files = [_os.path.join(directory, i) for i in _os.listdir(directory)]
    files = [i for i in files if _os.path.isfile(i)]

    if not files:
        _logme.log('No files found.', 'debug')
        return []

    deleted = []
    for f in files:
        for extension in extensions:
            if f.endswith(extension):
                deleted.append(f)

    deleted = sorted(deleted)
    delete  = False

    if confirm:
        if deleted:
            prompt = [_os.path.basename(i) for i in deleted]
            _sys.stdout.write('Directory: {}\n'.format(directory))
            _sys.stdout.write('Files to delete::\n\t')
            _sys.stdout.write('\n\t'.join(prompt) + '\n')
            answer = _run.get_input("Do you want to delete these files? [Y/n]",
                                    'yesno', 'y')
            if answer == 'y':
                delete  = True
                _sys.stdout.write('Deleting...\n')
            else:
                _sys.stdout.write('Aborting\n')
                delete  = False
                deleted = []
        else:
            _sys.stdout.write('No files to delete.\n')
    else:
        delete = True

    if delete and deleted:
        for f in deleted:
            _os.remove(f)
        if confirm:
            _sys.stdout.write('Done\n')

    return deleted



# -*- coding: utf-8 -*-
"""
A library of useful functions used throughout the *fyrd* package.

These include functions to handle data, format outputs, handle file opening,
run commands, check file extensions, get user input, and search and format
imports.

These functions are not intended to be accessed directly.
"""
from __future__ import print_function
from __future__ import with_statement
import os as _os
import re as _re
import sys as _sys
import inspect as _inspect
import argparse as _argparse

import bz2
import gzip
from subprocess import Popen
from subprocess import PIPE
from time import sleep

from six import reraise
from six.moves import input as _get_input

from . import logme as _logme


###############################################################################
#                               Useful Classes                                #
###############################################################################


class CustomFormatter(_argparse.ArgumentDefaultsHelpFormatter,
                      _argparse.RawDescriptionHelpFormatter):

    """Custom argparse formatting."""

    pass


class CommandError(Exception):

    """A custom exception."""

    pass


###############################################################################
#                               Misc Functions                                #
###############################################################################


def listify(iterable):
    """Try to force any iterable into a list sensibly."""
    if isinstance(iterable, list):
        return iterable
    if isinstance(iterable, (str, int, float)):
        return [iterable]
    if not iterable:
        return []
    if callable(iterable):
        iterable = iterable()
    return list(iter(iterable))


def merge_lists(lists):
    """Turn a list of lists into a single list."""
    outlist = []
    for lst in listify(lists):
        outlist += lst
    return outlist


def write_iterable(iterable, outfile):
    """Write all elements of iterable to outfile."""
    with open_zipped(outfile, 'w') as fout:
        fout.write('\n'.join(iterable))


def indent(string, prefix='    '):
    """Replicate python3's textwrap.indent for python2.

    Args:
        string (str): Any string.
        prefix (str): What to indent with.

    Returns:
        str: Indented string
    """
    out = ''
    for i in string.split('\n'):
        out += '{}{}\n'.format(prefix, i)
    return out


def is_exc(x):
    """Check if x is the output of sys.exc_info().

    Returns:
        bool: True if matched the output of sys.exc_info().
    """
    return bool(isinstance(x, tuple)
                and len(x) == 3
                and issubclass(BaseException, x[0]))


###############################################################################
#                               File Management                               #
###############################################################################


def open_zipped(infile, mode='r'):
    """Open a regular, gzipped, or bz2 file.

    If infile is a file handle or text device, it is returned without
    changes.

    Returns:
        text mode file handle.
    """
    mode   = mode[0] + 't'
    if hasattr(infile, 'write'):
        return infile
    if isinstance(infile, str):
        if infile.endswith('.gz'):
            return gzip.open(infile, mode)
        if infile.endswith('.bz2'):
            if hasattr(bz2, 'open'):
                return bz2.open(infile, mode)
            else:
                return bz2.BZ2File(infile, mode)
        return open(infile, mode)


def block_read(files, size=65536):
    """Iterate through a file by blocks."""
    while True:
        b = files.read(size)
        if not b:
            break
        yield b


def count_lines(infile, force_blocks=False):
    """Return the line count of a file as quickly as possible.

    Uses `wc` if avaialable, otherwise does a rapid read.
    """
    if which('wc') and not force_blocks:
        _logme.log('Using wc', 'debug')
        if infile.endswith('.gz'):
            cat = 'zcat'
        elif infile.endswith('.bz2'):
            cat = 'bzcat'
        else:
            cat = 'cat'
        command = "{cat} {infile} | wc -l | awk '{{print $1}}'".format(
            cat=cat, infile=infile
        )
        return int(cmd(command)[1])
    else:
        _logme.log('Using block read', 'debug')
        with open_zipped(infile) as fin:
            return sum(bl.count("\n") for bl in block_read(fin))


def split_file(infile, parts, outpath='', keep_header=False):
    """Split a file in parts and return a list of paths.

    NOTE: Linux specific (uses wc).

    **Note**: If has_header is True, the top line is stripped off the infile
    prior to splitting and assumed to be the header.

    Args:
        outpath:     The directory to save the split files.
        has_header:  Add the header line to the top of every file.

    Returns:
        list: Paths to split files.
    """
    # Determine how many reads will be in each split sam file.
    _logme.log('Getting line count', 'debug')

    num_lines = int(count_lines(infile)/int(parts)) + 1

    # Subset the file into X number of jobs, maintain extension
    cnt       = 0
    currjob   = 1
    suffix    = '.split_' + str(currjob).zfill(4) + '.' + infile.split('.')[-1]
    file_name = _os.path.basename(infile)
    run_file  = _os.path.join(outpath, file_name + suffix)
    outfiles  = [run_file]

    # Actually split the file
    _logme.log('Splitting file', 'debug')
    with open_zipped(infile) as fin:
        header = fin.readline() if keep_header else ''
        sfile = open_zipped(run_file, 'w')
        sfile.write(header)
        for line in fin:
            cnt += 1
            if cnt < num_lines:
                sfile.write(line)
            elif cnt == num_lines:
                sfile.write(line)
                sfile.close()
                currjob += 1
                suffix = '.split_' + str(currjob).zfill(4) + '.' + \
                    infile.split('.')[-1]
                run_file = _os.path.join(outpath, file_name + suffix)
                sfile = open_zipped(run_file, 'w')
                outfiles.append(run_file)
                sfile.write(header)
                cnt = 0
        sfile.close()
    _logme.log('Split files: {}'.format(outfiles), 'debug')
    return tuple(outfiles)


def is_exe(fpath):
    """Return True is fpath is executable."""
    return _os.path.isfile(fpath) and _os.access(fpath, _os.X_OK)


def file_type(infile):
    """Return file type after stripping gz or bz2."""
    name_parts = infile.split('.')
    if name_parts[-1] == 'gz' or name_parts[-1] == 'bz2':
        name_parts.pop()
    return name_parts[-1]


def is_file_type(infile, types):
    """Return True if infile is one of types.

    Args:
        infile:  Any file name
        types:   String or list/tuple of strings (e.g ['bed', 'gtf'])

    Returns:
        True or False
    """
    if hasattr(infile, 'write'):
        infile = infile.name
    types = listify(types)
    for typ in types:
        if file_type(infile) == typ:
            return True
    return False


###############################################################################
#                              Running Commands                               #
###############################################################################


def cmd(command, args=None, stdout=None, stderr=None, tries=1):
    """Run command and return status, output, stderr.

    Args:
        command (str): Path to executable.
        args (tuple):  Tuple of arguments.
        stdout (str):  File or open file like object to write STDOUT to.
        stderr (str):  File or open file like object to write STDERR to.
        tries (int):   Number of times to try to execute. 1+

    Returns:
        tuple: exit_code, STDOUT, STDERR
    """
    tries = int(tries)
    assert tries > 0
    count = 1
    if isinstance(command, (list, tuple)):
        if args:
            raise ValueError('Cannot submit list/tuple command as ' +
                             'well as args argument')
        command = ' '.join(command)
    assert isinstance(command, str)
    if args:
        if isinstance(args, (list, tuple)):
            args = ' '.join(args)
        args = command + args
    else:
        args = command
    _logme.log('Running {} as {}'.format(command, args), 'verbose')
    while True:
        try:
            pp = Popen(args, shell=True, universal_newlines=True,
                       stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            _logme.log('{} does not exist'.format(command), 'critical')
            raise
        out, err = pp.communicate()
        code = pp.returncode
        if code == 0 or count == tries:
            break
        _logme.log('Command {} failed with code {}, retrying.'
                   .format(command, code), 'warn')
        sleep(1)
        count += 1
    _logme.log('{} completed with code {}'.format(command, code), 'debug')
    if stdout:
        with open_zipped(stdout, 'w') as fout:
            fout.write(out)
    if stderr:
        with open_zipped(stderr, 'w') as fout:
            fout.write(err)
    return code, out.rstrip(), err.rstrip()


def export_run(function, args, kwargs):
    """Execute a function after first exporting all imports."""
    kwargs['imports'] = export_imports(function, kwargs)
    print('bob', kwargs['imports'])
    return function(*args, **kwargs)


def which(program):
    """Replicate the UNIX which command.

    Taken verbatim from:
        stackoverflow.com/questions/377017/test-if-executable-exists-in-python

    Args:
        program: Name of executable to test.

    Returns:
        Path to the program or None on failu_re.
    """
    fpath, program = _os.path.split(program)
    if fpath:
        if is_exe(program):
            return _os.path.abspath(program)
    else:
        for path in _os.environ["PATH"].split(_os.pathsep):
            path = path.strip('"')
            exe_file = _os.path.join(path, program)
            if is_exe(exe_file):
                return _os.path.abspath(exe_file)

    return None


def check_pid(pid):
    """Check For the existence of a unix pid."""
    try:
        _os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


###############################################################################
#                       Option and Argument Management                        #
###############################################################################


def replace_argument(args, find_string, replace_string, error=True):
    """Replace find_string with replace string in a tuple or dict.

    If dict, the values are replaced, not the keys.

    Note: args can also be a list, in which case the first item is assumed
    to be a tuple, and the second a dictionary

    Args:
        args (list/tuple/dict): Tuple or dict of args
        find_string (str):      A string to search for
        replace_string (str):   A string to replace with
        error (bool):           Raise ValueError if replacement fails

    Returns:
        The same object as was passed, with alterations made.
    """
    double = False
    if isinstance(args, list):
        args, kwargs = args
        double = True
    elif isinstance(args, tuple):
        kwargs = None
    elif isinstance(args, dict):
        kwargs = args.copy()
        args   = None
    else:
        raise ValueError('args must be list/tuple/dict, is {}\nval: {}'
                         .format(type(args), args))

    if not args and not kwargs:
        msg = 'No arguments or keyword arguments found'
        if error:
            raise ValueError(msg)
        else:
            _logme.log(msg, 'warn')
            if double:
                return None, None
            else:
                return None

    found = False
    newargs = tuple()
    if args:
        for arg in listify(args):
            if isinstance(arg, str) and find_string in arg:
                arg = arg.format(**{find_string.strip('{}'): replace_string})
                found = True
            newargs += (arg,)
    newkwds = {}
    if kwargs:
        for arg, value in kwargs.items():
            if isinstance(value, str) and find_string in value:
                value = replace_string
                found = True
            newkwds[arg] = value

    if found is not True:
        msg = 'Could not find {}'.format(find_string)
        if error:
            raise ValueError(msg)
        else:
            _logme.log(msg, 'warn')
            if double:
                return None, None
            else:
                return None

    if double:
        return [newargs, newkwds]
    else:
        if newargs:
            return newargs
        else:
            return newkwds


def opt_split(opt, split_on):
    """Split options by chars in split_on, merge all into single list.

    Args:
        opt (list):      A list of strings, can be a single string.
        split_on (list): A list of characters to use to split the options.

    Returns:
        list: A single merged list of split options, uniqueness guaranteed,
              order not.
    """
    opt = listify(opt)
    split_on = listify(split_on)
    final_list = []
    for o in opt:
        final_list += _re.split('[{}]'.format(''.join(split_on)), o)
    return list(set(final_list)) # Return unique options only, order lost.


###############################################################################
#                                 User Input                                  #
###############################################################################


def get_yesno(message, default=None):
    """Get yes/no answer from user.

    Args:
        message (str): A message to print, an additional space will be added.
        default (str): One of {'y', 'n'}, the default if the user gives no
                       answer. If None, answer forced.

    Returns:
        bool: True on yes, False on no
    """
    if default:
        if default.lower().startswith('y'):
            tailstr = '[Y/n] '
        elif default.lower().startswith('n'):
            tailstr = '[y/N] '
        else:
            raise ValueError('Invalid default')
    else:
        tailstr = '[y/n] '
    message = message + tailstr if message.endswith(' ') \
                else message + ' ' + tailstr
    ans = get_input(message, 'yesno', default)
    if ans.lower().startswith('y'):
        return True
    elif ans.lower().startswith('n'):
        return False
    else:
        raise ValueError('Invalid response: {}'.format(ans))


def get_input(message, valid_answers=None, default=None):
    """Get input from the command line and check answers.

    Allows input to work with python 2/3

    Args:
        message (str):        A message to print, an additional space will be
                              added.
        valid_answers (list): A list of answers to accept, if None, ignored.
                              Case insensitive. There is one special option
                              here: 'yesno', this allows all case insensitive
                              variations of y/n/yes/no.
        default (str):        The default answer.

    Returns:
        str: The response
    """
    if not message.endswith(' '):
        message = message + ' '
    if valid_answers:
        if isinstance(valid_answers, str):
            if valid_answers.lower() == 'yesno':
                valid_answers = ['yes', 'no', 'y', 'n']
            else:
                valid_answers = [valid_answers]
        if not isinstance(valid_answers, (list, tuple, set, frozenset)):
            _logme.log('valid_answers must be a list, is {}'
                       .format(type(valid_answers)), 'critical')
            raise ValueError('Invalid argument')
        valid_answers = [i.lower() for i in valid_answers]
        while True:
            ans = _get_input(message)
            if not ans and default:
                return default
            if ans.lower() in valid_answers:
                return ans
            else:
                _logme.log('Invalid response to input question', 'debug')
                _sys.stderr.write('Invalid response: {}\n'.format(ans) +
                                  'Valid responses: {}\n'
                                  .format(valid_answers) +
                                  'Please try again.\n')
    else:
        return _get_input(message)


###############################################################################
#                                   Imports                                   #
###############################################################################


def syspath_fmt(syspaths):
    """Take a list of paths and return a sys of sys.path.append strings."""
    outlist = []
    for pth in listify(syspaths):
        if 'sys.path' in pth:
            outlist.append(pth)
            continue
        if _os.path.exists(pth):
            outlist.append("sys.path.append('{}')".format(
                _os.path.abspath(pth)
            ))
        else:
            raise OSError('Paths must exist, {} does not.'
                          .format(pth))
    return '\n'.join(outlist)


PROT_IMPT = """\
try:
    {}
except ImportError:
    pass
"""


def normalize_imports(imports, prot=True):
    """Take a heterogenous list of imports and normalize it.

    Args:
        imports (list): A list of strings, formatted differently.
        prot (bool):    Protect imports with try..except blocks

    Returns:
        list: A list of strings that can be used for imports
    """
    out_impts  = []
    prot_impts = []
    path_impts = []
    imports = listify(imports)
    if not imports:
        return []
    for imp in imports:
        if not isinstance(imp, str):
            raise ValueError('All imports must be strings')
        if imp.startswith('try:'):
            prot_impts.append(imp.rstrip())
        elif imp.startswith('import') or imp.startswith('from'):
            out_impts.append(imp.rstrip())
        elif imp.startswith('sys.path.append')\
            or imp.startswith('sys.path.insert'):
            path_impts.append(imp.rstrip())
        else:
            if imp.startswith('@'):
                continue
            out_impts.append('import {}'.format(imp))

    if prot:
        for imp in out_impts:
            prot_impts.append(PROT_IMPT.format(imp))
        out = prot_impts
    else:
        out = out_impts + prot_impts

    # Remove duplicates
    out = list(set(out))

    # Add PATHS
    if path_impts:
        out = list(set(path_impts)) + out

    return out


def get_function_path(function):
    """Return path to module defining a function if it exists."""
    mod = _inspect.getmodule(function)
    if mod and mod != '__main__':
        return _os.path.dirname(_inspect.getabsfile(function))
    else:
        return None


def update_syspaths(function, kwds=None):
    """Add function path to 'syspaths' in kwds."""
    if kwds:
        syspaths = listify(kwds['syspaths']) if 'syspaths' in kwds else []
    else:
        syspaths = []
    return [get_function_path(function)] + syspaths


def import_function(function, mode='string'):
    """Return an import string for the function.

    Attempts to resolve the parent module also, if the parent module is a file,
    ie it isn't __main__, the import string will include a call to
    sys.path.append to ensure the module is importable.

    If this function isn't defined by a module, returns an empty string.

    Args:
        mode (str): string/list, return as a unified string or a list.
    """
    if not callable(function):
        raise ValueError('Function must be callable, {} is not'
                         .format(function))
    if mode not in ['string', 'list']:
        raise ValueError("Invalid mode {}, must be 'list' or 'string'"
                         .format(mode))

    if _inspect.ismethod(function):
        name = (dict(_inspect.getmembers(function.__self__))['__class__']
                .__name__)
    else:
        name    = function.__name__

    # Attempt to resolve defining file
    parent = _inspect.getmodule(function)

    imports = []
    if parent and parent.__name__ != '__main__':
        path   = _os.path.dirname(parent.__file__)
        module = parent.__name__
        # If module is the child of a package, change the directory up to the
        # parent
        if '.' in module:
            path = _os.path.abspath(
                _os.path.join(
                    path, *['..' for i in range(module.count('.'))]
                )
            )
        imports.append("sys.path.append('{}')".format(path))
        imports.append('import {}'.format(module))
        imports.append('from {} import *'.format(module))
        imports.append('from {} import {}'.format(module, name))

    return imports if mode == 'list' else '\n'.join(imports)


def get_imports(function, mode='string'):
    """Build a list of potentially useful imports from a function handle.

    Gets:

        - All modules from globals()
        - All modules from the function's globals()
        - All functions from the function's globals()

    Modes:
        string: Return a list of strings formatted as unprotected import calls
        prot:   Similar to string, but with try..except blocks
        list:   Return two lists: (import name, module name) for modules
                and (import name, function name, module name) for functions

    Args:
        function (callable): A function handle
        mode (str):          A string corresponding to one of the above modes

    Returns:
        str or list
    """
    if mode not in ['string', 'prot', 'list']:
        raise ValueError('mode must be one of string/prot/list')

    rootmod  = _inspect.getmodule(function)

    imports      = []
    func_imports = []

    # For interactive sessions
    members = dict(_inspect.getmembers(function))
    locations = [members]
    if '__globals__' in members:
        locations.append(members['__globals__'])
    for location in locations:
        for name, item in location.items():
            if name.startswith('__'):
                continue
            # Modules
            if _inspect.ismodule(item):
                imports.append((name, item.__name__))
            # Functions
            elif callable(item):
                try:
                    func_imports.append((name, item.__name__, item.__module__))
                except AttributeError:
                    pass

    # Import all modules in the root module
    imports += [(k,v.__name__)
                for k,v in _inspect.getmembers(rootmod, _inspect.ismodule)
                if not k.startswith('__')]

    # Make unique
    imports      = sorted(list(set(imports)), key=_sort_imports)
    func_imports = sorted(list(set(func_imports)), key=_sort_imports)

    _logme.log('Imports: {}'.format(imports), 'debug')
    _logme.log('Func imports: {}'.format(func_imports), 'debug')

    # Create a sane set of imports
    ignore_list = ['os', 'sys', 'dill', 'pickle', '__main__']
    filtered_imports      = []
    filtered_func_imports = []
    for iname, name in imports:
        if iname in ignore_list:
            continue
        if name.startswith('@') or iname.startswith('@'):
            continue
        filtered_imports.append((iname, name))
    for iname, name, mod in func_imports:
        if iname in ignore_list:
            continue
        if name.startswith('@') or iname.startswith('@'):
            continue
        filtered_func_imports.append((iname, name, mod))

    if mode == 'list':
        return filtered_imports, filtered_func_imports

    import_strings = []
    for iname, name in filtered_imports:
        names = name.split('.')
        if names[0] == '__main__':
            continue
        if iname != name:
            if len(names) > 1:
                if '.'.join(names[1:]) != iname:
                    import_strings.append(
                        'from {} import {} as {}'
                        .format('.'.join(names[:-1]), names[-1], iname)
                    )
                else:
                    import_strings.append(
                        'from {} import {}'
                        .format(names[0], '.'.join(names[1:]))
                    )
            else:
                import_strings.append(
                    ('import {} as {}').format(name, iname)
                )
        else:
            import_strings.append('import {}'.format(name))

    # Function imports
    for iname, name, mod in filtered_func_imports:
        if mod == '__main__':
            continue
        if iname == name:
            import_strings.append('from {} import {}'.format(mod, name))
        else:
            import_strings.append('from {} import {} as {}'
                                  .format(mod, name, iname)
                                 )

    if mode == 'string':
        return import_strings

    elif mode == 'prot':
        return normalize_imports(import_strings, prot=True)

    else:
        raise ValueError('Mode changed unexpectedly')


def export_globals(function):
    """Add a function's globals to the current globals."""
    rootmod = _inspect.getmodule(function)
    globals()[rootmod.__name__] = rootmod
    for k, v in _inspect.getmembers(rootmod, _inspect.ismodule):
        if not k.startswith('__'):
            globals()[k] = v


def get_all_imports(function, kwds, prot=False):
    """Get all imports from a function and from kwds.

    Args:
        function (callable): A function handle
        kwds (dict):         A dictionary of keyword arguments
        prot (bool):         Wrap all import in try statement

    Returns:
        list: Imports
    """
    imports  = listify(kwds['imports'] if 'imports' in kwds else None)
    imports  = normalize_imports(imports, prot=False)
    imports += get_imports(function, mode='string')
    return normalize_imports(imports, prot=prot)


def export_imports(function, kwds):
    """Get imports from a function and from kwds.

    Also sets globals and adds path to module to sys path.

    Args:
        function (callable): A function handle
        kwds (dict):         A dictionary of keyword arguments

    Returns:
        list: imports + sys.path.append for module path
    """
    export_globals(function)
    return import_function(function, 'list') + get_all_imports(function, kwds)


def _sort_imports(x):
    """Sort a list of tuples and strings, for use with sorted."""
    if isinstance(x, tuple):
        if x[1] == '__main__':
            return 0
        return x[1]
    return x
